#
# Copyright (c), 2018-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
"""
XPath 2.0 implementation - part 2 (operators, expressions and multi-role tokens)
"""
import math
import operator
from copy import copy
from collections.abc import Iterator
from decimal import Decimal, DivisionByZero
from typing import cast, NoReturn

import elementpath.aliases as ta

from elementpath.protocols import XsdAttributeProtocol
from elementpath.exceptions import ElementPathError
from elementpath.namespaces import XSD_NAMESPACE, XSD_NOTATION, XSD_ANY_ATOMIC_TYPE, XSD_UNTYPED
from elementpath.helpers import numeric_equal, numeric_not_equal, \
    node_position, get_double
from elementpath.namespaces import XSD_ERROR, get_namespace, get_expanded_name
from elementpath.datatypes import UntypedAtomic, QName, AnyURI, \
    Duration, Integer, DoubleProxy10
from elementpath.xpath_nodes import ElementNode, DocumentNode, XPathNode, AttributeNode
from elementpath.sequences import xlist
from elementpath.sequence_types import is_instance
from elementpath.xpath_context import XPathSchemaContext
from elementpath.xpath_tokens import XPathToken, XPathFunction, XPathConstructor

from .xpath2_parser import XPath2Parser

__all__ = ['XPath2Parser']

COMPARISON_OPERATORS = frozenset(('eq', 'ne', 'lt', 'le', 'gt', 'ge'))

register = XPath2Parser.register
infix = XPath2Parser.infix
method = XPath2Parser.method
function = XPath2Parser.function


@method('then')
@method('as')
@method('of')
@method('else')
@method('in')
@method('return')
@method('satisfies')
def nud__auxiliary_symbols(self: XPathToken) -> XPathToken:
    return self.as_name()


###
# Node sequence composition
XPath2Parser.duplicate('|', 'union')


@method(infix('intersect', bp=55))
@method(infix('except', bp=55))
def select__intersect_and_except_operators(self: XPathToken, context: ta.ContextType = None) \
        -> Iterator[XPathNode]:
    if context is None:
        raise self.missing_context()

    s1, s2 = set(self[0].select(copy(context))), set(self[1].select(copy(context)))
    if any(not isinstance(x, XPathNode) for x in s1) \
            or any(not isinstance(x, XPathNode) for x in s2):
        raise self.error('XPTY0004', 'only XPath nodes are allowed')

    if self.symbol == 'except':
        yield from cast(list[XPathNode], sorted(s1 - s2, key=node_position))
    else:
        yield from cast(list[XPathNode], sorted(s1 & s2, key=node_position))


###
# 'if' expression
@method('if', bp=20, label='expression')
def nud__if_expression(self: XPathToken) -> XPathToken:
    if self.parser.next_token.symbol != '(':
        return self.as_name()

    self.parser.advance('(')
    self[:] = self.parser.expression(5),
    self.parser.advance(')')
    self.parser.advance('then')
    self[1:] = self.parser.expression(5),
    self.parser.advance('else')
    self[2:] = self.parser.expression(5),
    return self


@method('if')
def evaluate__if_expression(self: XPathToken, context: ta.ContextType = None) \
        -> ta.ValueType:
    if self.boolean_value(self[0].select(copy(context))):
        if isinstance(context, XPathSchemaContext):
            self[2].evaluate(context)
        return self[1].evaluate(context)
    else:
        if isinstance(context, XPathSchemaContext):
            self[1].evaluate(context)
        return self[2].evaluate(context)


@method('if')
def select__if_expression(self: XPathToken, context: ta.ContextType = None) \
        -> Iterator[ta.ItemType]:
    if self.boolean_value(self[0].select(copy(context))):
        if isinstance(context, XPathSchemaContext):
            self[2].evaluate(context)
        yield from self[1].select(context)
    else:
        if isinstance(context, XPathSchemaContext):
            self[1].evaluate(context)
        yield from self[2].select(context)


###
# Quantified expressions
@method('some', bp=20, label='expression')
@method('every', bp=20, label='expression')
def nud__quantified_expressions(self: XPathToken) -> XPathToken:
    del self[:]
    if self.parser.next_token.symbol != '$':
        return self.as_name()

    while True:
        self.parser.next_token.expected('$')
        variable = self.parser.expression(5)
        self.append(variable)
        self.parser.advance('in')
        expr = self.parser.expression(5)
        self.append(expr)
        for tk in filter(lambda x: x.symbol == '$', expr.iter()):
            if tk[0].value == variable[0].value:
                raise tk.error('XPST0008', 'loop variable in its range expression')

        if self.parser.next_token.symbol != ',':
            break
        self.parser.advance()

    self.parser.advance('satisfies')
    self.append(self.parser.expression(5))
    return self


@method('some')
@method('every')
def evaluate__quantified_expressions(self: XPathToken, context: ta.ContextType = None) -> bool:
    if context is None:
        raise self.missing_context()

    context = copy(context)
    context.variables = context.variables.copy()

    some = self.symbol == 'some'
    varnames = [cast(str, self[k][0].value) for k in range(0, len(self) - 1, 2)]
    selectors = [self[k].select for k in range(1, len(self) - 1, 2)]

    for results in context.iter_product(selectors, varnames):
        context.variables.update(x for x in zip(varnames, results))
        if self.boolean_value(self[-1].select(copy(context))):
            if some:
                return True
        elif not some:
            return False

    return not some


###
# 'for' expressions
@method('for', bp=20, label='expression')
def nud__for_expression(self: XPathToken) -> XPathToken:
    del self[:]
    if self.parser.next_token.symbol != '$':
        return self.as_name()

    while True:
        self.parser.next_token.expected('$')
        variable = self.parser.expression(5)
        self.append(variable)
        self.parser.advance('in')
        expr = self.parser.expression(5)
        self.append(expr)
        for tk in filter(lambda x: x.symbol == '$', expr.iter()):
            if tk[0].value == variable[0].value:
                raise tk.error('XPST0008', 'loop variable in its range expression')

        if self.parser.next_token.symbol != ',':
            break
        self.parser.advance()

    self.parser.advance('return')
    self.append(self.parser.expression(5))
    return self


@method('for')
def select__for_expression(self: XPathToken, context: ta.ContextType = None) \
        -> Iterator[ta.ItemType]:
    if context is None:
        raise self.missing_context()

    context = copy(context)
    context.variables = context.variables.copy()

    varnames = [cast(str, self[k][0].value) for k in range(0, len(self) - 1, 2)]
    selectors = [self[k].select for k in range(1, len(self) - 1, 2)]

    for results in copy(context).iter_product(selectors, varnames):
        context.variables.update(x for x in zip(varnames, results))
        yield from self[-1].select(copy(context))


###
# Sequence type based
@method('instance', bp=60, label='expression')
@method('treat', bp=61, label='expression')
def led__sequence_type_based_expressions(self: XPathToken, left: XPathToken) -> XPathToken:
    self.parser.advance('of' if self.symbol == 'instance' else 'as')
    self[:] = left, self.parser.parse_sequence_type()
    return self


@method('instance')
def evaluate__instance_expression(self: XPathToken, context: ta.ContextType = None) -> bool:
    occurs = self[1].occurrence
    position = None

    if self[1].symbol == 'empty-sequence':
        for _ in self[0].select(context):
            return False
        return True
    elif self[1].label in ('kind test', 'sequence type', 'function test'):
        if context is None:
            raise self.missing_context()

        for position, context.item in enumerate(self[0].select(context)):
            if context.axis is None:
                context.axis = 'self'

            result = self[1].evaluate(context)
            if isinstance(result, list) and not result:
                return occurs in ('*', '?') or \
                    isinstance(context.item, XPathFunction) and \
                    context.item.name == XSD_ERROR
            elif position and occurs in ('', '?'):
                return False
        else:
            return position is not None or occurs in ('*', '?')
    else:
        type_name = self[1].source.rstrip('*+?')
        try:
            qname = get_expanded_name(type_name, self.parser.namespaces)
        except KeyError as err:
            raise self.error('XPST0081', "namespace prefix {} not found".format(err))

        for position, item in enumerate(self[0].select(context)):
            try:
                if not is_instance(item, qname, self.parser):
                    return False
            except TypeError as err:
                raise self.error('XPST0051', err) from None
            except KeyError:
                msg = f"atomic type {type_name!r} not found in in-scope schema types"
                raise self.error('XPST0051', msg) from None
            else:
                if position and occurs in ('', '?'):
                    return False
        else:
            return position is not None or occurs in ('*', '?')


@method('treat')
def evaluate__treat_expression(self: XPathToken, context: ta.ContextType = None) \
        -> ta.ValueType:
    occurs = self[1].occurrence
    position = None
    castable_expr: xlist[ta.ItemType] = xlist()
    if self[1].symbol == 'empty-sequence':
        for _ in self[0].select(context):
            raise self.error('XPDY0050')
    elif self[1].label in ('kind test', 'sequence type', 'function test'):
        for position, item in enumerate(self[0].select(context)):
            result = self[1].evaluate(context)
            if not result and isinstance(result, list):
                raise self.error('XPDY0050')
            elif position and occurs in ('', '?'):
                raise self.error('XPDY0050', "more than one item in sequence")
            castable_expr.append(item)
        else:
            if position is None and occurs not in ('*', '?'):
                raise self.error('XPDY0050', "the sequence cannot be empty")
    else:
        type_name = self[1].source.rstrip('*+?')
        try:
            qname = get_expanded_name(type_name, self.parser.namespaces)
        except KeyError as err:
            raise self.error('XPST0081', 'prefix {} not found'.format(str(err)))

        if not qname.startswith('{') and not QName.is_valid(qname):
            raise self.error('XPST0003')

        for position, item in enumerate(self[0].select(context)):
            try:
                if not is_instance(item, qname, self.parser):
                    msg = f"item {item!r} is not of type {type_name!r}"
                    raise self.error('XPDY0050', msg)
            except KeyError:
                msg = f"atomic type {type_name!r} not found in in-scope schema types"
                raise self.error('XPST0051', msg) from None
            else:
                if position and occurs in ('', '?'):
                    raise self.error('XPDY0050', "more than one item in sequence")
                castable_expr.append(item)
        else:
            if position is None and occurs not in ('*', '?'):
                raise self.error('XPDY0050', "the sequence cannot be empty")

    return castable_expr


###
# Simple type based expressions

@method('castable', bp=62, label='expression')
@method('cast', bp=63, label='expression')
def led__cast_expressions(self: XPathToken, left: XPathToken) -> XPathToken:
    self.parser.advance('as')
    self.parser.expected_next('(name)', ':', 'Q{', message='an EQName expected')
    self[:] = left, self.parser.expression(rbp=85)
    if self.parser.next_token.symbol == '?':
        self[1].occurrence = '?'
        self.parser.advance()
    return self


@method('castable')
@method('cast')
def evaluate__cast_expressions(self: XPathToken, context: ta.ContextType = None) \
        -> ta.OneAtomicOrEmpty:
    type_name = self[1].source.rstrip('+*?')
    try:
        atomic_type = get_expanded_name(type_name, self.parser.namespaces)
    except KeyError as err:
        raise self.error('XPST0081', 'prefix {} not found'.format(str(err)))

    if atomic_type in (XSD_NOTATION, XSD_ANY_ATOMIC_TYPE):
        raise self.error('XPST0080')

    namespace = get_namespace(atomic_type)
    if namespace != XSD_NAMESPACE and \
            (self.parser.schema is None or self.parser.schema.get_type(atomic_type) is None):
        msg = f"atomic type {atomic_type!r} not found in the in-scope schema types"
        raise self.error('XPST0051', msg)

    result = [res for res in self[0].select(context)]
    if len(result) > 1:
        if self.symbol != 'cast':
            return False
        raise self.error('XPTY0004', "more than one value in expression")
    elif not result:
        if self[1].occurrence == '?':
            return [] if self.symbol == 'cast' else True
        elif self.symbol != 'cast':
            return False
        else:
            raise self.error('XPTY0004', "an atomic value is required")

    arg = self.data_value(result[0])
    value: ta.OneAtomicOrEmpty
    try:
        if namespace != XSD_NAMESPACE:
            if self.parser.schema is not None:
                value = self.parser.schema.cast_as(self.string_value(arg), atomic_type)
            else:
                value = []
        else:
            local_name = atomic_type.split('}')[1]
            try:
                token_class = cast(type[XPathConstructor], self.parser.symbol_table[local_name])
            except KeyError:
                msg = f"atomic type {type_name!r} not found in the in-scope schema types"
                raise self.error('XPST0051', msg)
            else:
                if token_class.label != 'constructor function':
                    msg = f"token {type_name!r} is not a constructor"
                    raise self.error('XPST0051', msg)

            if local_name == 'QName':
                if isinstance(arg, QName):
                    pass
                elif self.parser.version < '3.0' and self[0].symbol != '(string)':
                    raise self.error('XPTY0004', "Non literal string to QName cast")

            token = token_class(self.parser)
            value = token.cast(arg)

    except ElementPathError:
        if self.symbol != 'cast':
            return False
        elif isinstance(context, XPathSchemaContext):
            return UntypedAtomic('1')
        raise
    except (TypeError, ValueError) as err:
        if self.symbol != 'cast':
            return False
        elif isinstance(context, XPathSchemaContext):
            return UntypedAtomic('1')
        elif isinstance(arg, (UntypedAtomic, str)):
            raise self.error('FORG0001', err) from None
        raise self.error('XPTY0004', err) from None
    else:
        return value if self.symbol == 'cast' else True


###
# Comma operator - concatenate items or sequences
@method(infix(',', bp=5))
def evaluate__comma_operator(self: XPathToken, context: ta.ContextType = None) \
        -> list[ta.ItemType]:
    results: list[ta.ItemType] = []
    for op in self:
        result = op.evaluate(context)
        if isinstance(result, list):
            results.extend(result)
        elif result is not None:
            results.append(result)
    return results


@method(',')
def select__comma_operator(self: XPathToken, context: ta.ContextType = None) \
        -> Iterator[ta.ItemType]:
    for op in self:
        yield from op.select(copy(context))


###
# Parenthesized expression: XPath 2.0 admits the empty case ().
@method(register('(', lbp=80, rpb=80, label='expression'))
def nud__parenthesized_expression(self: XPathToken) -> XPathToken:
    if self.parser.next_token.symbol != ')':
        self[:] = self.parser.expression(),
    self.parser.advance(')')
    return self


@method('(')
def led__parenthesized_expression(self: XPathToken, left: XPathToken) -> XPathToken:
    if left.symbol == '(name)':
        if left.value in self.parser.RESERVED_FUNCTION_NAMES:
            msg = f"{left.value!r} is not allowed as function name"
            raise left.error('XPST0003', msg)
        else:
            raise left.error('XPST0017', 'unknown function {!r}'.format(left.value))

    elif left.symbol == ':' and left[1].symbol == '(name)':
        if left[1].namespace == XSD_NAMESPACE:
            msg = 'unknown constructor function {!r}'.format(left[1].value)
            raise left[1].error('XPST0017', msg)
        raise left.error('XPST0017', 'unknown function {!r}'.format(left.value))

    if self.parser.next_token.symbol != ')':
        self[:] = left, self.parser.expression()
    else:
        self[:] = left,
    self.parser.advance(')')
    return self


@method('(')
def evaluate__parenthesized_expression(self: XPathToken, context: ta.ContextType = None) \
        -> ta.ValueType:
    return self[0].evaluate(context) if self else []


@method('(')
def select__parenthesized_expression(self: XPathToken, context: ta.ContextType = None) \
        -> Iterator[ta.ItemType]:
    return self[0].select(context) if self else iter(())


###
# Value comparison operators (eq, ne, lt, le, gt, and ge)
#
# Ref: https://www.w3.org/TR/xpath20/#id-value-comparisons
#
@method('eq', bp=30)
@method('ne', bp=30)
@method('lt', bp=30)
@method('gt', bp=30)
@method('le', bp=30)
@method('ge', bp=30)
def led__value_comparison_operators(self: XPathToken, left: XPathToken) -> XPathToken:
    if left.symbol in COMPARISON_OPERATORS:
        raise self.wrong_syntax()
    self[:] = left, self.parser.expression(rbp=30)
    return self


@method('eq')
@method('ne')
@method('lt')
@method('gt')
@method('le')
@method('ge')
def evaluate__value_comparison_operators(self: XPathToken, context: ta.ContextType = None) \
        -> ta.OneOrEmpty[bool]:
    operands = [self[0].get_atomized_operand(context), self[1].get_atomized_operand(context)]

    if any(x is None for x in operands):
        return []
    elif any(isinstance(x, XPathFunction) for x in operands):
        raise self.error('FOTY0013', "cannot compare a function item")
    elif all(isinstance(x, DoubleProxy10) for x in operands):
        # Special case of two <class 'float'> values: use custom operators
        if self.symbol == 'eq':
            return numeric_equal(*cast(list[float], operands))
        elif self.symbol == 'ne':
            return numeric_not_equal(*cast(list[float], operands))
        elif numeric_equal(*cast(list[float], operands)):
            return self.symbol in ('le', 'ge')

    cls0, cls1 = type(operands[0]), type(operands[1])
    if cls0 is cls1 and cls0 is not Duration:
        pass
    elif all(isinstance(x, float) for x in operands):
        pass
    elif any(isinstance(x, bool) for x in operands):
        msg = "cannot apply {} between {!r} and {!r}".format(self, *operands)
        raise self.error('XPTY0004', msg)
    elif all(isinstance(x, (int, Decimal)) for x in operands):
        pass
    elif all(isinstance(x, (str, UntypedAtomic, AnyURI)) for x in operands):
        pass
    elif all(isinstance(x, (str, UntypedAtomic, QName)) for x in operands):
        pass
    elif all(isinstance(x, (float, Decimal, int)) for x in operands):
        if isinstance(operands[0], float):
            operands[1] = get_double(cast(ta.NumericType, operands[1]), self.parser.xsd_version)
        else:
            operands[0] = get_double(cast(ta.NumericType, operands[0]), self.parser.xsd_version)
    elif all(isinstance(x, Duration) for x in operands) and self.symbol in ('eq', 'ne'):
        pass
    elif (issubclass(cls0, cls1) or issubclass(cls1, cls0)) and not issubclass(cls0, Duration):
        pass
    else:
        msg = "cannot apply {} between {!r} and {!r}".format(self, *operands)
        raise self.error('XPTY0004', msg)

    try:
        return cast(bool, getattr(operator, self.symbol)(*operands))
    except TypeError as err:
        raise self.error('XPTY0004', err) from None


###
# Node comparison
@method('is', bp=30)
def led__node_comparison(self: XPathToken, left: XPathToken) -> XPathToken:
    if left.symbol == 'is':
        raise self.wrong_syntax()
    self[:] = left, self.parser.expression(rbp=30)
    return self


@method('is')
@method(infix('<<', bp=30))
@method(infix('>>', bp=30))
def evaluate__node_comparison(self: XPathToken, context: ta.ContextType = None) \
        -> ta.OneOrEmpty[bool]:
    symbol = self.symbol

    left = [x for x in self[0].select(context)]
    if not left:
        return []
    elif len(left) > 1 or not isinstance(left[0], XPathNode):
        raise self[0].error('XPTY0004', f"left operand of {symbol!r} must be a single node")

    right = [x for x in self[1].select(context)]
    if not right:
        return []
    elif len(right) > 1 or not isinstance(right[0], XPathNode):
        raise self[0].error('XPTY0004', "right operand of %r must be a single node" % symbol)

    if symbol == 'is':
        return left[0] is right[0]
    else:
        if left[0] is right[0] or context is None:
            return False

        documents = [context.root]
        documents.extend(v for v in context.variables.values() if isinstance(v, DocumentNode))

        for root in documents:
            if root is not None:
                for item in root.iter_document():  # pragma: no cover
                    if left[0] is item:
                        return True if symbol == '<<' else False
                    elif right[0] is item:
                        return False if symbol == '<<' else True
        else:
            raise self.error('FOCA0002', "operands are not nodes of the XML tree!")


###
# Range expression
@method('to', bp=35)
def led__range_expression(self: XPathToken, left: XPathToken) -> XPathToken:
    if left.symbol == 'to':
        raise self.wrong_syntax()
    self[:] = left, self.parser.expression(rbp=35)
    return self


@method('to')
def evaluate__range_expression(self: XPathToken, context: ta.ContextType = None) -> list[int]:
    start, stop = self.get_operands(context, cls=Integer)
    try:
        return xlist(range(start, stop + 1))
    except TypeError:
        return []


@method('to')
def select__range_expression(self: XPathToken, context: ta.ContextType = None) -> Iterator[int]:
    yield from cast(list[int] | list[NoReturn], self.evaluate(context))


###
# Numerical operators
@method(infix('idiv', bp=45))
def evaluate__idiv_operator(self: XPathToken, context: ta.ContextType = None) -> int:
    op1, op2 = self.get_operands(context)
    if op1 is None or op2 is None:
        raise self.error('XPST0005')

    try:
        if math.isinf(op1):
            raise self.error('FOAR0001' if op2 == 0 else 'FOAR0002')
        elif math.isnan(op1) or math.isnan(op2):
            raise self.error('FOAR0002')
    except TypeError as err:
        if isinstance(context, XPathSchemaContext):
            return 1
        raise self.error('XPTY0004', err) from None

    try:
        result = op1 // op2
    except (ZeroDivisionError, DivisionByZero):
        if isinstance(context, XPathSchemaContext):
            return 1
        raise self.error('FOAR0001') from None
    else:
        if result >= 0 or isinstance(op1, Decimal) or \
                isinstance(op2, Decimal) or abs(op1) == abs(op2):
            return int(result)
        else:
            return int(result) + 1


# Resolve the intrinsic ambiguity of some infix operators
@method('union')
@method('intersect')
@method('except')
@method('eq')
@method('ne')
@method('lt')
@method('gt')
@method('le')
@method('ge')
@method('is')
@method('to')
@method('idiv')
@method('instance')
@method('treat')
@method('castable')
@method('cast')
def nud__disambiguation_of_infix_operators(self: XPathToken) -> XPathToken:
    return self.as_name()


###
# Kind tests (sequence types that can appear also in XPath expressions)
@method(function('document-node', nargs=(0, 1), label='kind test'))
def select__document_node_kind_test(self: XPathFunction, context: ta.ContextType = None) \
        -> Iterator[DocumentNode]:
    if context is None:
        raise self.missing_context()
    elif not self:
        for item in context.iter_children_or_self():
            if isinstance(item, DocumentNode):
                yield item
    else:
        elements = [e for e in self[0].select(copy(context)) if isinstance(e, ElementNode)]
        if isinstance(context.item, DocumentNode):
            if len(elements) == 1:
                yield context.item


@method('document-node')
def nud__document_node_kind_test(self: XPathFunction) -> XPathFunction:
    self.parser.advance('(')
    if self.parser.next_token.symbol in ('element', 'schema-element'):
        self[0:] = self.parser.expression(5),
        if self.parser.next_token.symbol == ',':
            msg = 'Too many arguments: expected at most 1 argument'
            raise self.error('XPST0017', msg)
    elif self.parser.next_token.symbol != ')':
        raise self.error('XPST0003', 'element or schema-element kind test expected')
    self.parser.advance(')')
    return self


@method(function('element', nargs=(0, 2), label='kind test'))
def select__element_kind_test(self: XPathFunction, context: ta.ContextType = None) \
        -> Iterator[ElementNode]:
    if context is None:
        raise self.missing_context()
    elif not self:
        for item in context.iter_children_or_self():
            if isinstance(item, ElementNode):
                yield item
    else:
        for item in self[0].select(context):
            if len(self) == 1:
                yield cast(ElementNode, item)  # Already selected by sequence type test
            elif isinstance(item, ElementNode):
                type_annotation = self[1].name
                if item.nilled:
                    if self[1].occurrence in ('*', '?'):
                        yield item
                elif item.type_name == type_annotation:
                    if type_annotation != XSD_UNTYPED:
                        yield item
                    elif self[0].symbol != '*':
                        yield item
                elif is_instance(item.typed_value, type_annotation, self.parser):
                    yield item


@method('element')
def nud__element_kind_test(self: XPathFunction) -> XPathFunction:
    self.parser.advance('(')
    if self.parser.next_token.symbol != ')':
        self.parser.expected_next('(name)', ':', '*', message='a QName or a wildcard expected')
        self[0:] = self.parser.expression(5),
        if self.parser.next_token.symbol == ',':
            self.parser.advance(',')
            self.parser.expected_next('(name)', ':', message='a QName expected')
            self[1:] = self.parser.expression(80),
            if self.parser.next_token.symbol in ('*', '+', '?'):
                self[1].occurrence = self.parser.next_token.symbol
                self.parser.advance()

    self.parser.advance(')')
    return self


@method(function('schema-attribute', nargs=1, label='kind test'))
def select__schema_attribute_kind_test(self: XPathFunction, context: ta.ContextType = None) \
        -> Iterator[AttributeNode]:
    if context is None:
        raise self.missing_context()

    attribute_name = self[0].source
    qname = get_expanded_name(attribute_name, self.parser.namespaces)

    for _ in context.iter_children_or_self():
        if self.parser.schema is None:
            break

        if self.parser.schema.get_attribute(qname) is None:
            raise self.error('XPST0008', "attribute %r not found in schema" % attribute_name)

        if isinstance(context.item, AttributeNode) and context.item.match_name(qname):
            yield context.item
            return

    if not isinstance(context, XPathSchemaContext):
        raise self.error('XPST0008', 'schema attribute %r not found' % attribute_name)


@method(function('schema-element', nargs=1, label='kind test'))
def select__schema_element_kind_test(self: XPathFunction, context: ta.ContextType = None) \
        -> Iterator[ElementNode]:
    if context is None:
        raise self.missing_context()

    element_name = self[0].source
    qname = get_expanded_name(element_name, self.parser.namespaces)

    if self.parser.schema is not None:
        for _ in context.iter_children_or_self():
            if self.parser.schema.get_element(qname) is None:
                raise self.error('XPST0008', "element %r not found in schema" % element_name)

            if isinstance(context.item, ElementNode) and context.item.name == qname:
                yield context.item
                return

    if not isinstance(context, XPathSchemaContext):
        raise self.error('XPST0008', 'schema element %r not found' % element_name)


@method('schema-attribute')
@method('schema-element')
def nud__schema_node_kind_test(self: XPathFunction) -> XPathFunction:
    self.parser.advance('(')
    self.parser.expected_next('(name)', ':', 'Q{', message='a QName expected')
    self[0:] = self.parser.expression(5),
    self.parser.advance(')')
    return self


###
# Multi role-tokens definition: in XPath 2.0 the 'attribute' keyword is used both for
# attribute:: axis and attribute() node type function.
#
# First the XPath1 token class has to be removed from the XPath2 symbol table. Then the
# symbol has to be registered usually with the same binding power (bp --> lbp, rbp), a
# multi-value label (using a tuple of values) and a custom pattern. Finally a custom nud
# or led method is required.
XPath2Parser.unregister('attribute')
XPath2Parser.register(
    'attribute', lbp=90, rbp=90, label=('kind test', 'axis'),
    pattern=r'\battribute(?=\s*\:\:|\s*\(\:.*\:\)\s*\:\:|\s*\(|\s*\(\:.*\:\)\()'
)


@method('attribute')
def nud__attribute_kind_test_or_axis(self: XPathToken) -> XPathToken:
    if self.parser.next_token.symbol == '::':
        self.label = 'axis'
        self.parser.advance('::')
        self.parser.expected_next(
            '(name)', '*', 'text', 'node', 'document-node', 'comment', 'processing-instruction',
            'attribute', 'schema-attribute', 'element', 'schema-element', 'namespace-node'
        )
        self[:] = self.parser.expression(rbp=90),
    else:
        self.label = 'kind test'
        self.parser.advance('(')
        if self.parser.next_token.symbol != ')':
            self.parser.next_token.expected('(name)', '*', ':')
            self[:] = self.parser.expression(5),

            if self.parser.next_token.symbol == ',':
                self.parser.advance(',')
                self.parser.next_token.expected('(name)', ':')
                self[1:] = self.parser.expression(5),

        self.parser.advance(')')

        if self.namespace:
            msg = f"{self.value!r} is not allowed as function name"
            raise self.error('XPST0003', msg)

    return self


@method('attribute')
def select__attribute_kind_test_or_axis(self: XPathToken, context: ta.ContextType = None) \
        -> Iterator[ta.AtomicType | AttributeNode | XsdAttributeProtocol]:
    if context is None:
        raise self.missing_context()
    elif self.label == 'axis':
        for _ in context.iter_attributes():
            yield from cast(Iterator[AttributeNode], self[0].select(context))
    elif not self:
        for attribute in context.iter_attributes():
            yield attribute
    else:
        name = self[0].value
        assert isinstance(name, str)

        if self.parser.schema is not None and len(self) == 2:
            assert isinstance(self[1].value, str)
            type_name = get_expanded_name(self[1].value, namespaces=self.parser.namespaces)
        else:
            type_name = None

        for attribute in context.iter_attributes():
            if attribute.match_name(name):
                if isinstance(context, XPathSchemaContext):
                    continue

                if type_name == XSD_UNTYPED == attribute.type_name:
                    if name != '*':
                        yield attribute
                elif not type_name or attribute.type_name == type_name or \
                        is_instance(attribute.typed_value, type_name, self.parser):
                    yield attribute


# XPath 2.0 definitions continue into module xpath2_functions



#
# Copyright (c), 2018-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
"""
XPath 2.0 implementation - part 4 (XSD constructors)
"""
import decimal
from typing import cast, NoReturn

import elementpath.aliases as ta

from elementpath.exceptions import ElementPathError, ElementPathSyntaxError
from elementpath.namespaces import XSD_NAMESPACE
from elementpath.datatypes import AbstractDateTime, Duration, Date, DateTime, \
    DateTimeStamp, Time, UntypedAtomic, QName, HexBinary, Base64Binary, \
    BooleanProxy, AnyURI, Notation, NMToken, Idref, Entity, DateTime10, ErrorProxy
from elementpath.xpath_context import XPathSchemaContext
from elementpath.xpath_tokens import XPathConstructor
from elementpath.sequences import xlist
from ._xpath2_functions import XPath2Parser

register = XPath2Parser.register
unregister = XPath2Parser.unregister
method = XPath2Parser.method
constructor = XPath2Parser.constructor


@constructor('normalizedString')
@constructor('token')
@constructor('language')
@constructor('NMTOKEN')
@constructor('Name')
@constructor('NCName')
@constructor('ID')
@constructor('IDREF')
@constructor('ENTITY')
@constructor('anyURI')
def cast__string_types(self: XPathConstructor, value: ta.AtomicType) -> str | AnyURI:
    try:
        result = cast(str | AnyURI, self.type_class.make(value))
    except ValueError as err:
        raise self.error('FORG0001', err)
    else:
        assert isinstance(result, (str, AnyURI))
        return result


@constructor('decimal')
@constructor('double')
@constructor('float')
def cast__numeric_types(self: XPathConstructor, value: ta.AtomicType) -> ta.NumericType:
    try:
        result = self.type_class.make(value, parser=self.parser)
    except ValueError as err:
        if isinstance(value, (str, UntypedAtomic)):
            raise self.error('FORG0001', err)
        raise self.error('FOCA0002', err)
    except ArithmeticError as err:
        raise self.error('FOCA0002', err) from None
    else:
        assert isinstance(result, (int, float, decimal.Decimal))
        return result


@constructor('integer')
@constructor('nonNegativeInteger')
@constructor('positiveInteger')
@constructor('nonPositiveInteger')
@constructor('negativeInteger')
@constructor('long')
@constructor('int')
@constructor('short')
@constructor('byte')
@constructor('unsignedLong')
@constructor('unsignedInt')
@constructor('unsignedShort')
@constructor('unsignedByte')
def cast__integer_types(self: XPathConstructor, value: ta.AtomicType) -> int:
    try:
        return cast(int, self.type_class.make(value))
    except ValueError:
        msg = 'could not convert {!r} to xs:{}'.format(value, self.symbol)
        if isinstance(value, (str, bytes, int, UntypedAtomic)):
            raise self.error('FORG0001', msg) from None
        raise self.error('FOCA0002', msg) from None
    except ArithmeticError as err:
        raise self.error('FOCA0002', err) from None


@constructor('date')
@constructor('gDay')
@constructor('gMonth')
@constructor('gMonthDay')
@constructor('gYear')
@constructor('gYearMonth')
@constructor('time')
@constructor('dateTimeStamp')
def cast__other_datetime_types(self: XPathConstructor, value: ta.AtomicType) -> ta.DateTimeType:
    try:
        return cast(ta.DateTimeType, self.type_class).make(value, parser=self.parser)
    except OverflowError as err:
        raise self.error('FODT0001', err) from None
    except ValueError as err:
        raise self.error('FORG0001', err)


@method('date')
@method('gDay')
@method('gMonth')
@method('gMonthDay')
@method('gYear')
@method('gYearMonth')
@method('time')
def evaluate__other_datetime_types(self: XPathConstructor, context: ta.ContextType = None) \
        -> ta.OneOrEmpty[AbstractDateTime]:
    if self.context is not None:
        context = self.context

    arg = self.data_value(self.get_argument(context))
    if arg is None:
        return []

    try:
        return cast(AbstractDateTime, self.cast(arg))
    except (TypeError, OverflowError) as err:
        if isinstance(context, XPathSchemaContext):
            return []
        elif isinstance(err, TypeError):
            raise self.error('FORG0006', err) from None
        else:
            raise self.error('FODT0001', err) from None


@method('dateTimeStamp')
def evaluate__datetime_stamp_type(self: XPathConstructor, context: ta.ContextType = None) \
        -> ta.OneOrEmpty[DateTimeStamp]:
    if self.context is not None:
        context = self.context

    arg = self.data_value(self.get_argument(context))
    if arg is None:
        return []

    if isinstance(arg, UntypedAtomic):
        result = self.cast(arg.value)
    elif isinstance(arg, Date):
        result = self.cast(arg)
    else:
        result = self.cast(str(arg))
    assert isinstance(result, DateTimeStamp)
    return result


@method('dateTimeStamp')
def nud__datetime_stamp_type(self: XPathConstructor) -> XPathConstructor:
    if self.parser.xsd_version == '1.0':
        raise self.wrong_syntax("xs:dateTimeStamp is not recognized unless XSD 1.1 is enabled")
    if not self.parser.parse_arguments:
        return self

    try:
        self.parser.advance('(')
        self[0:] = self.parser.expression(5),
        if self.parser.next_token.symbol == ',':
            msg = 'Too many arguments: expected at most 1 argument'
            raise self.error('XPST0017', msg)
        self.parser.advance(')')
    except SyntaxError as err:
        raise self.error('XPST0017', str(err)) from None
    return self


@constructor('duration')
@constructor('yearMonthDuration')
@constructor('dayTimeDuration')
def cast__duration_types(self: XPathConstructor, value: ta.AtomicType) -> Duration:
    try:
        return cast(Duration, self.type_class).make(value)
    except OverflowError as err:
        raise self.error('FODT0002', err) from None
    except ValueError as err:
        raise self.error('FORG0001', err)


###
# Constructors for binary XSD types
@constructor('base64Binary')
@constructor('hexBinary')
def cast__binary_types(self: XPathConstructor, value: ta.AtomicType) -> Base64Binary | HexBinary:
    try:
        result = self.type_class.make(value, parser=self.parser)
    except ValueError as err:
        raise self.error('FORG0001', err) from None
    except TypeError as err:
        raise self.error('XPTY0004', err) from None
    else:
        assert isinstance(result, (Base64Binary, HexBinary))
        return result


@method('base64Binary')
@method('hexBinary')
def evaluate__binary_types(self: XPathConstructor, context: ta.ContextType = None) \
        -> ta.OneOrEmpty[HexBinary | Base64Binary]:
    arg = self.data_value(self.get_argument(self.context or context))
    if arg is None:
        return []

    try:
        return cast(HexBinary | Base64Binary, self.cast(arg))
    except ElementPathError as err:
        if isinstance(context, XPathSchemaContext):
            return []
        err.token = self
        raise


@constructor('NOTATION')
def cast__notation_type(self: XPathConstructor, value: ta.AtomicType) -> Notation:
    raise NotImplementedError("No value is castable to xs:NOTATION")


@method('NOTATION')
def nud__notation_type(self: XPathConstructor) -> None:
    if not self.parser.parse_arguments:
        return

    self.parser.advance('(')
    if self.parser.next_token.symbol == ')':
        raise self.error('XPST0017', 'expected exactly one argument')
    self[0:] = self.parser.expression(5),
    if self.parser.next_token.symbol != ')':
        raise self.error('XPST0017', 'expected exactly one argument')
    self.parser.advance()
    raise self.error('XPST0017', "no constructor function exists for xs:NOTATION")


###
# Multirole tokens (function or constructor function)
#

# Case 1: In XPath 2.0 the 'boolean' keyword is used both for boolean() function and
# for boolean() constructor function.
unregister('boolean')


@constructor('boolean', label=('function', 'constructor function'),
             sequence_types=('item()*', 'xs:boolean'))
def cast__boolean_type(self: XPathConstructor, value: ta.AtomicType) -> bool:
    try:
        return cast(bool, BooleanProxy(value))
    except ValueError as err:
        raise self.error('FORG0001', err) from None
    except TypeError as err:
        raise self.error('XPTY0004', err) from None


@method('boolean')
def nud__boolean_type_and_function(self: XPathConstructor) -> XPathConstructor:
    if not self.parser.parse_arguments:
        return self

    self.parser.advance('(')
    if self.parser.next_token.symbol == ')':
        msg = 'Too few arguments: expected at least 1 argument'
        raise self.error('XPST0017', msg)
    self[0:] = self.parser.expression(5),
    if self.parser.next_token.symbol == ',':
        msg = 'Too many arguments: expected at most 1 argument'
        raise self.error('XPST0017', msg)
    self.parser.advance(')')
    return self


@method('boolean')
def evaluate__boolean_type_and_function(self: XPathConstructor, context: ta.ContextType = None) \
        -> ta.OneOrEmpty[bool]:
    if self.context is not None:
        context = self.context

    if self.label == 'function':
        return self.boolean_value(self[0].select(context))

    # xs:boolean constructor
    arg = self.data_value(self.get_argument(context))
    if arg is None:
        return []

    try:
        return cast(bool, self.cast(arg))
    except ElementPathError as err:
        if isinstance(context, XPathSchemaContext):
            return []
        err.token = self
        raise


###
# Case 2: In XPath 2.0 the 'string' keyword is used both for fn:string() and xs:string().
unregister('string')


@constructor('string', label=('function', 'constructor function'),
             nargs=(0, 1), sequence_types=('item()?', 'xs:string'))
def cast__string_type(self: XPathConstructor, value: ta.AtomicType) -> str:
    return self.string_value(value)


@method('string')
def nud__string_type_and_function(self: XPathConstructor) -> XPathConstructor:
    if not self.parser.parse_arguments:
        return self

    try:
        self.parser.advance('(')
        if self.label != 'function' or self.parser.next_token.symbol != ')':
            self[0:] = self.parser.expression(5),
        self.parser.advance(')')
    except ElementPathSyntaxError as err:
        raise self.error('XPST0017', err)
    else:
        return self


@method('string')
def evaluate__string_type_and_function(self: XPathConstructor, context: ta.ContextType = None) \
        -> ta.OneOrEmpty[str]:
    if self.context is not None:
        context = self.context

    if self.label == 'function':
        if not self:
            if context is None:
                raise self.missing_context()
            return self.string_value(context.item)
        return self.string_value(self.get_argument(context))
    else:
        item = self.get_argument(context)
        return [] if item is None else self.string_value(item)


# Case 3 and 4: In XPath 2.0 the XSD 'QName' and 'dateTime' types have special
# constructor functions so the 'QName' keyword is used both for fn:QName() and
# xs:QName(), the same for 'dateTime' keyword.
#
# In those cases the label at parse time is set by the nud method, in dependence
# of the number of args.
#
@constructor('QName', bp=90, label=('function', 'constructor function'),
             nargs=(1, 2), sequence_types=('xs:string?', 'xs:string', 'xs:QName'))
def cast__qname_type(self: XPathConstructor, value: str) -> QName:
    return self.cast_to_qname(value)


@constructor('dateTime', bp=90, label=('function', 'constructor function'),
             nargs=(1, 2), sequence_types=('xs:date?', 'xs:time?', 'xs:dateTime?'))
def cast__datetime_type(self: XPathConstructor, value: ta.AtomicType) -> DateTime | None:
    try:
        result = cast(DateTime, self.type_class).make(value, parser=self.parser)
    except OverflowError as err:
        raise self.error('FODT0001', err) from None
    except ValueError as err:
        raise self.error('FORG0001', err) from None
    else:
        return result


@method('QName')
@method('dateTime')
def nud__qname_and_datetime(self: XPathConstructor) -> XPathConstructor:
    if not self.parser.parse_arguments:
        return self

    try:
        self.parser.advance('(')
        self[0:] = self.parser.expression(5),
        if self.parser.next_token.symbol == ',':
            if self.label != 'function':
                raise self.error('XPST0017', 'unexpected 2nd argument')
            self.label = 'function'
            self.parser.advance(',')
            self[1:] = self.parser.expression(5),
        elif self.label != 'constructor function' or self.namespace != XSD_NAMESPACE:
            raise self.error('XPST0017', '2nd argument missing')
        else:
            self.label = 'constructor function'
            self.nargs = 1
        self.parser.advance(')')
    except SyntaxError:
        raise self.error('XPST0017') from None
    else:
        return self


@method('QName')
def evaluate__qname_type_and_function(self: XPathConstructor, context: ta.ContextType = None) \
        -> ta.OneOrEmpty[QName]:
    if self.context is not None:
        context = self.context

    if self.label == 'constructor function':
        arg = self.data_value(self.get_argument(context))
        if arg is None:
            return []
        value = self.cast(arg)
        assert isinstance(value, QName)
        return value
    else:
        uri = self.get_argument(context)
        qname = self.get_argument(context, index=1)
        try:
            return QName(uri, qname)
        except (TypeError, ValueError) as err:
            if isinstance(context, XPathSchemaContext):
                return []
            elif isinstance(err, TypeError):
                raise self.error('XPTY0004', err)
            else:
                raise self.error('FOCA0002', err)


@method('dateTime')
def evaluate__datetime_type_and_function(self: XPathConstructor, context: ta.ContextType = None) \
        -> ta.OneOrEmpty[DateTime]:
    if self.context is not None:
        context = self.context

    if self.label == 'constructor function':
        arg = self.data_value(self.get_argument(context))
        if arg is None:
            return []

        try:
            result = self.cast(arg)
        except (ValueError, TypeError) as err:
            if isinstance(context, XPathSchemaContext):
                return []
            elif isinstance(err, ValueError):
                raise self.error('FORG0001', err) from None
            else:
                raise self.error('FORG0006', err) from None
        else:
            assert isinstance(result, DateTime)
            return result
    else:
        dt = self.get_argument(context, cls=Date)
        tm = self.get_argument(context, 1, cls=Time)
        if dt is None or tm is None:
            return []
        elif dt.tzinfo == tm.tzinfo or tm.tzinfo is None:
            tzinfo = dt.tzinfo
        elif dt.tzinfo is None:
            tzinfo = tm.tzinfo
        else:
            raise self.error('FORG0008')

        cls = DateTime if self.parser.xsd_version == '1.1' else DateTime10
        return cls(dt.year, dt.month, dt.day, tm.hour, tm.minute,
                   tm.second, tm.microsecond, tzinfo)


@constructor('untypedAtomic')
def cast__untyped_atomic(self: XPathConstructor, value: ta.AtomicType) -> UntypedAtomic:
    return UntypedAtomic(value)


@method('untypedAtomic')
def evaluate__untyped_atomic(self: XPathConstructor, context: ta.ContextType = None) \
        -> ta.OneOrEmpty[UntypedAtomic]:
    arg = self.data_value(self.get_argument(self.context or context))
    if arg is None:
        return []
    elif isinstance(arg, UntypedAtomic):
        return arg
    else:
        arg = self.cast(arg)
        assert isinstance(arg, UntypedAtomic)
        return arg


###
# The fn:error function and the xs:error constructor.
#
# https://www.w3.org/TR/2010/REC-xpath-functions-20101214/#func-error
# https://www.w3.org/TR/xpath-functions/#func-error
#
#

# TODO: apply sequence_types=('xs:anyAtomicType?', 'xs:error?') for xs:error
@constructor('error', bp=90, label=('function', 'constructor function'), nargs=(0, 3),
             sequence_types=('xs:QName?', 'xs:string', 'item()*', 'none'))
def cast__error_type(self: XPathConstructor, value: ta.AtomicType) -> list[NoReturn]:
    try:
        result = ErrorProxy(value)
    except (TypeError, ValueError) as err:
        raise self.error('FORG0001', str(err)) from None
    else:
        if result is None:
            return []
        elif not isinstance(result, list) or len(result):
            raise self.error('FORG0001')
        return result


@method('error')
def nud__error_type_and_function(self: XPathConstructor) -> XPathConstructor:
    self.clear()
    if not self.parser.parse_arguments:
        return self

    try:
        self.parser.advance('(')
        if self.namespace == XSD_NAMESPACE:
            self.label = 'constructor function'
            self.nargs = 1
            if self.parser.xsd_version == '1.0':
                raise self.error('XPST0051', 'xs:error is not defined with XSD 1.0')
            self.append(self.parser.expression(5))
        else:
            self.label = 'function'
            for k in range(3):
                if self.parser.next_token.symbol == ')':
                    break
                self.append(self.parser.expression(5))
                if self.parser.next_token.symbol == ')':
                    break
                self.parser.advance(',')
        self.parser.advance(')')
    except SyntaxError:
        raise self.error('XPST0017') from None
    else:
        return self


@method('error')
def evaluate__error_type_and_function(self: XPathConstructor,
                                      context: ta.ContextType = None) -> None:
    if self.context is not None:
        context = self.context

    error: QName | None
    if self.label == 'constructor function':
        self.cast(self.get_argument(context))
    elif not self:
        raise self.error('FOER0000')
    elif len(self) == 1:
        error = self.get_argument(context, cls=QName)
        if error is None and self.parser.version <= '3.0':
            raise self.error('XPTY0004', "an xs:QName expected")
        raise self.error(error or 'FOER0000')
    else:
        error = self.get_argument(context, cls=QName)
        description: str | None = self.get_argument(context, index=1, cls=str)
        raise self.error(error or 'FOER0000', description)


###
# XSD list-based constructors

@constructor('NMTOKENS', sequence_types=('xs:NMTOKEN*',))
def cast__nmtokens(self: XPathConstructor, value: ta.AtomicType) -> list[NMToken]:
    if isinstance(value, UntypedAtomic):
        values = value.value.split() or [value.value]
    elif hasattr(value, 'split'):
        values = value.split() or [value]
    else:
        raise self.error('FORG0001')

    try:
        return xlist([NMToken(x) for x in values])
    except ValueError as err:
        raise self.error('FORG0001', err) from None


@constructor('IDREFS', sequence_types=('xs:IDREF*',))
def cast__idrefs(self: XPathConstructor, value: ta.AtomicType) -> list[Idref]:
    if isinstance(value, UntypedAtomic):
        values = value.value.split() or [value.value]
    elif hasattr(value, 'split'):
        values = value.split() or [value]
    else:
        raise self.error('FORG0001')

    try:
        return xlist([Idref(x) for x in values])
    except ValueError as err:
        raise self.error('FORG0001', err) from None


@constructor('ENTITIES', sequence_types=('xs:ENTITY*',))
def cast__entities(self: XPathConstructor, value: ta.AtomicType) -> list[Entity]:
    if isinstance(value, UntypedAtomic):
        values = value.value.split() or [value.value]
    elif hasattr(value, 'split'):
        values = value.split() or [value]
    else:
        raise self.error('FORG0001')

    try:
        return xlist([Entity(x) for x in values])
    except ValueError as err:
        raise self.error('FORG0001', err) from None



#
# Copyright (c), 2016-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
import re
from sys import maxunicode
from collections import Counter
from collections.abc import Callable, Iterator, MutableSet
from itertools import chain
from typing import AbstractSet, Any, Optional, Union

from .codepoints import RegexError
from .unicode_subsets import UnicodeSubset, lazy_subset, unicode_subset, unicode_category

I_SHORTCUT_REPLACE = (
    ":A-Z_a-z\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u02FF\u0370-\u037D\u037F-\u1FFF"
    "\u200C-\u200D\u2070-\u218F\u2C00-\u2FEF\u3001-\uD7FF\uF900-\uFDCF\uFDF0-\uFFFD"
)

C_SHORTCUT_REPLACE = (
    "-.0-9:A-Z_a-z\u00B7\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u037D\u037F-\u1FFF\u200C-"
    "\u200D\u203F\u2040\u2070-\u218F\u2C00-\u2FEF\u3001-\uD7FF\uF900-\uFDCF\uFDF0-\uFFFD"
)


@lazy_subset
def c_shortcut() -> UnicodeSubset:
    return UnicodeSubset(C_SHORTCUT_REPLACE)


@lazy_subset
def i_shortcut() -> UnicodeSubset:
    return UnicodeSubset(I_SHORTCUT_REPLACE)


@lazy_subset
def s_shortcut() -> UnicodeSubset:
    return UnicodeSubset(' \t\n\r')


@lazy_subset
def d_shortcut() -> UnicodeSubset:
    return unicode_category('Nd')


@lazy_subset
def w_shortcut() -> UnicodeSubset:
    return UnicodeSubset(chain.from_iterable(unicode_category(x) for x in 'LMNS'))


# Single and Multi character escapes
CHARACTER_ESCAPES: dict[str, Union[str, Callable[[], UnicodeSubset]]] = {
    # Single-character escapes
    '\\n': '\n',
    '\\r': '\r',
    '\\t': '\t',
    '\\|': '|',
    '\\.': '.',
    '\\-': '-',
    '\\^': '^',
    '\\?': '?',
    '\\*': '*',
    '\\+': '+',
    '\\{': '{',
    '\\}': '}',
    '\\(': '(',
    '\\)': ')',
    '\\[': '[',
    '\\]': ']',
    '\\\\': '\\',

    # Multi-character escapes
    '\\s': s_shortcut,
    '\\S': s_shortcut,
    '\\d': d_shortcut,
    '\\D': d_shortcut,
    '\\i': i_shortcut,
    '\\I': i_shortcut,
    '\\c': c_shortcut,
    '\\C': c_shortcut,
    '\\w': w_shortcut,
    '\\W': w_shortcut,
}


class CharacterClass(MutableSet[int]):
    """
    A set class to represent XML Schema/XQuery/XPath regex character class.

    :param charset: a string with formatted character set.
    :param xsd_version: the reference XSD version for syntax variants. Defaults to '1.0'.
    TODO: implement __ior__, __iand__, __ixor__ operators for a full mutable set class.
    """
    _re_char_set = re.compile(r'(?<!.-)(\\[nrt|.\-^?*+{}()\]sSdDiIcCwW]|\\[pP]{[a-zA-Z\-0-9]+})')
    _re_unicode_ref = re.compile(r'\\([pP]){([\w-]+)}')

    __slots__ = 'xsd_version', 'positive', 'negative'

    def __init__(self, charset: Optional[str] = None, xsd_version: str = '1.0') -> None:
        self.xsd_version = xsd_version
        self.positive = UnicodeSubset()
        self.negative = UnicodeSubset()
        if charset:
            self.add(charset)

    def __repr__(self) -> str:
        return '%s(%s)' % (self.__class__.__name__, str(self))

    def __str__(self) -> str:
        if not self.negative:
            return '[%s]' % str(self.positive)
        elif not self.positive:
            return '[^%s]' % str(self.negative)
        else:
            return '[%s%s]' % (
                str(UnicodeSubset(self.negative.complement())), str(self.positive)
            )

    def __copy__(self) -> 'CharacterClass':
        obj = CharacterClass(xsd_version=self.xsd_version)
        obj.positive.update(self.positive)
        obj.negative.update(self.negative)
        return self

    def __contains__(self, item: object) -> bool:
        if isinstance(item, str):
            item = ord(item)
        elif not isinstance(item, int):
            return False

        if self.negative:
            return item not in self.negative or item in self.positive
        return item in self.positive

    def __iter__(self) -> Iterator[int]:
        if self.negative:
            return (
                cp for cp in range(maxunicode + 1)
                if cp in self.positive or cp not in self.negative
            )
        return iter(sorted(self.positive))  # type: ignore[arg-type]

    def __len__(self) -> int:
        if self.negative:
            not_in_positive = Counter(x not in self.positive for x in self.negative)[True]
            return maxunicode + 1 - not_in_positive
        return len(self.positive)

    def __isub__(self, other: AbstractSet[Any]) -> 'CharacterClass':
        if isinstance(other, CharacterClass):
            if self.negative:
                if other.negative:
                    self.positive |= (other.negative - self.negative)
                    self.negative.clear()
                self.negative |= other.positive
            elif other.negative:
                self.positive &= other.negative
            self.positive -= other.positive
            return self
        return NotImplemented

    def __sub__(self, other: AbstractSet[Any]) -> 'CharacterClass':
        obj = self.__copy__()
        return obj.__isub__(other)

    def add(self, charset: Union[int, str]) -> None:
        if isinstance(charset, int):
            charset = chr(charset)

        for part in self._re_char_set.split(charset):
            if part in CHARACTER_ESCAPES:
                value = CHARACTER_ESCAPES[part]
                if isinstance(value, str):
                    self.positive.update(value)
                elif part[-1].islower():
                    self.positive |= value()
                else:
                    self.negative |= value()
            elif part.startswith('\\p') or part.startswith('\\P'):
                if self._re_unicode_ref.search(part) is None:
                    raise RegexError("wrong Unicode block specification %r" % part)

                try:
                    subset = unicode_subset(part[3:-1])
                except RegexError:
                    # XSD 1.1 supports Is prefix to match Unicode blocks
                    if not self.xsd_version or not part[3:].startswith('Is'):
                        raise
                    self.positive |= UnicodeSubset([(0, maxunicode + 1)])
                else:
                    if part.startswith('\\p'):
                        self.positive |= subset
                    else:
                        self.negative |= subset
            else:
                self.positive.update(part)

    def discard(self, charset: Union[int, str]) -> None:
        if isinstance(charset, int):
            charset = chr(charset)

        for part in self._re_char_set.split(charset):
            if part in CHARACTER_ESCAPES:
                value = CHARACTER_ESCAPES[part]
                if isinstance(value, str):
                    self.positive.difference_update(value)
                    if self.negative:
                        self.negative.update(value)
                elif part[-1].islower():
                    self.positive -= value()
                    if self.negative:
                        self.negative |= value()
                else:
                    self.positive &= value()
                    self.negative.clear()

            elif part.startswith('\\p') or part.startswith('\\P'):
                if self._re_unicode_ref.search(part) is None:
                    raise RegexError("wrong Unicode block specification %r" % part)

                try:
                    subset = unicode_subset(part[3:-1])
                except RegexError:
                    # XSD 1.1 supports Is prefix to match Unicode blocks
                    if not self.xsd_version or not part[3:].startswith('Is'):
                        raise
                    self.positive -= UnicodeSubset([(0, maxunicode + 1)])
                else:
                    if part.startswith('\\p'):
                        self.positive -= subset
                    else:
                        self.negative -= subset
            else:
                self.positive.difference_update(part)

    def clear(self) -> None:
        self.positive.clear()
        self.negative.clear()

    def complement(self) -> None:
        if self.positive or self.negative:
            self.positive, self.negative = self.negative, self.positive
        else:
            self.positive.codepoints = [(0, maxunicode + 1)]



#
# Copyright (c), 2018-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
from abc import ABCMeta, abstractmethod
from types import MappingProxyType
from typing import Any

from elementpath.helpers import LazyPattern
from elementpath.namespaces import XSD_NAMESPACE

###
# Classes for XSD built-in atomic types. All defined classes use a
# metaclass that adds some common methods and registers each class
# into a dictionary. Some classes of XSD primitive types are defined
# as proxies of basic Python datatypes.

_builtin_atomic_types: dict[str, 'AtomicTypeMeta'] = {}
builtin_atomic_types = MappingProxyType(_builtin_atomic_types)
"""Registry of builtin atomic types by expanded name and prefixed name."""


class AtomicTypeMeta(ABCMeta):
    """
    Metaclass for creating builtin atomic types. The created classes
    are decorated with missing attributes and methods. When a name
    attribute is provided the class is registered into two maps of
    atomic referenced by expanded name and prefixed name. For default
    the names are mapped to XSD namespace with 'xs' as the prefix.
    """
    types_map = _builtin_atomic_types

    def __new__(mcs, class_name: str, bases: tuple[type[Any], ...],
                dict_: dict[str, Any]) -> 'AtomicTypeMeta':
        try:
            name = dict_['name']
        except KeyError:
            name = dict_['name'] = None  # do not inherit name

        if name is not None and not isinstance(name, str):
            raise TypeError("attribute 'name' must be a string or None")

        if '__slots__' not in dict_:
            dict_['__slots__'] = ()  # Avoids __dict__ creation

        cls = super(AtomicTypeMeta, mcs).__new__(mcs, class_name, bases, dict_)

        # Register all the derived classes with a valid name if not already registered
        if name:
            namespace: str | None = dict_.pop('namespace', XSD_NAMESPACE)
            prefix: str | None = dict_.pop('prefix', 'xs')

            extended_name = f'{{{namespace}}}{name}' if namespace else name
            prefixed_name = f'{prefix}:{name}' if prefix else name

            if not cls.__doc__:
                if ver := getattr(cls, '_xsd_version', ''):
                    ver = ver + ' '
                params = getattr(cls, '__init__', None).__doc__ or ''
                cls.__doc__ = f"Class for XSD{ver} {prefixed_name} builtin datatype.\n{params}"

            if extended_name not in cls.types_map:
                assert prefixed_name not in mcs.types_map, \
                    'prefix already used by another namespace'
                mcs.types_map[extended_name] = mcs.types_map[prefixed_name] = cls

        return cls


class AnyType(metaclass=ABCMeta):
    """
    xs:anyType: the root type for all XML Schema types. Not definable as a concrete type by XPath.
    """

    name: str | None = None          # Register the class if a name is provided.
    namespace: str | None = None     # Namespace URI, defaults to XSD namespace.
    prefix: str | None = None        # Namespace prefix: defaults to 'xs' prefix.
    _xsd_version: str | None = None  # Overridden by types compatible with either XSD 1.0 or 1.1

    @abstractmethod
    def __init__(self, obj: Any) -> None:
        raise NotImplementedError()

    @classmethod
    def make(cls, value: Any, *args: Any, **kwargs: Any) -> 'AnyType | float':
        """
        Versioned factory method to create XSD type instances.

        :param value: the value to be converted to the XSD type.
        :param kwargs: additional keyword arguments to be passed to factory.
        """
        return cls(value)

    @classmethod
    def validate(cls, value: Any) -> None:
        """Validate the provided value against the XSD type."""
        if not isinstance(value, object):
            raise cls._invalid_type(value)  # noqa

    @classmethod
    def is_valid(cls, value: Any) -> bool:
        """Return whether the provided value is a valid XSD type."""
        try:
            cls.validate(value)
        except (TypeError, ValueError):
            return False
        else:
            return True

    @classmethod
    def _invalid_type(cls, obj: object) -> TypeError:
        if cls.name:
            return TypeError('invalid type {!r} for xs:{}'.format(type(obj), cls.name))
        return TypeError('invalid type {!r} for {!r}'.format(type(obj), cls))

    @classmethod
    def _invalid_value(cls, obj: object) -> ValueError:
        if cls.name:
            return ValueError('invalid value {!r} for xs:{}'.format(obj, cls.name))
        return ValueError('invalid value {!r} for {!r}'.format(obj, cls))

    @property
    def xpath_versions(self) -> tuple[str, ...] | str:
        """Return the XPath versions that the instance is compatible with."""
        return '2.0+'

    @property
    def xsd_versions(self) -> tuple[str, str] | str:
        """Return the XSD versions that the instance is compatible with."""
        return self._xsd_version or ('1.0', '1.1')


class AnySimpleType(AnyType):
    """
    xs:anySimpleType: the base type of all XSD simple types (atomic types,
    union types and list types). Not applicable as concrete type by XPath.
    """


class AnyAtomicType(AnySimpleType, metaclass=AtomicTypeMeta):
    """xs:anyAtomicType: the base type of all XSD atomic types."""
    name = 'anyAtomicType'
    pattern = LazyPattern(r'^$')

    @abstractmethod
    def __init__(self, value: Any) -> None:
        raise NotImplementedError()

    @classmethod
    def validate(cls, value: object) -> None:
        if isinstance(value, cls):
            return
        elif isinstance(value, str):
            if cls.pattern.match(value) is None:
                raise cls._invalid_value(value)
        else:
            raise cls._invalid_type(value)



#
# Copyright (c), 2018-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
"""
Subpackage for processing XML regular expressions and for converting them to
Python-compatible regexps.

XPath/XQuery/XML-Schema regexp flavors are supported through translate_pattern()
API options. Default options process XPath/XQuery patterns.
"""
from .codepoints import RegexError, iter_code_points
from .unicode_subsets import UnicodeSubset, UnicodeData, install_unicode_data, \
    unicode_version, unicode_subset, lazy_subset, unicode_category, unicode_block
from .character_classes import CharacterClass
from .patterns import translate_pattern

__all__ = ['translate_pattern', 'RegexError', 'UnicodeSubset', 'UnicodeData',
           'install_unicode_data', 'unicode_version', 'unicode_subset', 'lazy_subset',
           'unicode_category', 'unicode_block', 'CharacterClass', 'iter_code_points']



#
# Copyright (c), 2018-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or https://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
# --- Auto-generated code: don't edit this file ---
#
UNICODE_BLOCKS_VER_2_0_0 = {
    'Basic Latin': '\u0000-\u007F',
    'Latin-1 Supplement': '\u0080-\u00FF',
    'Latin Extended-A': '\u0100-\u017F',
    'Latin Extended-B': '\u0180-\u024F',
    'IPA Extensions': '\u0250-\u02AF',
    'Spacing Modifier Letters': '\u02B0-\u02FF',
    'Combining Diacritical Marks': '\u0300-\u036F',
    'Greek': '\u0370-\u03FF',
    'Cyrillic': '\u0400-\u04FF',
    'Armenian': '\u0530-\u058F',
    'Hebrew': '\u0590-\u05FF',
    'Arabic': '\u0600-\u06FF',
    'Devanagari': '\u0900-\u097F',
    'Bengali': '\u0980-\u09FF',
    'Gurmukhi': '\u0A00-\u0A7F',
    'Gujarati': '\u0A80-\u0AFF',
    'Oriya': '\u0B00-\u0B7F',
    'Tamil': '\u0B80-\u0BFF',
    'Telugu': '\u0C00-\u0C7F',
    'Kannada': '\u0C80-\u0CFF',
    'Malayalam': '\u0D00-\u0D7F',
    'Thai': '\u0E00-\u0E7F',
    'Lao': '\u0E80-\u0EFF',
    'Tibetan': '\u0F00-\u0FBF',
    'Georgian': '\u10A0-\u10FF',
    'Hangul Jamo': '\u1100-\u11FF',
    'Latin Extended Additional': '\u1E00-\u1EFF',
    'Greek Extended': '\u1F00-\u1FFF',
    'General Punctuation': '\u2000-\u206F',
    'Superscripts and Subscripts': '\u2070-\u209F',
    'Currency Symbols': '\u20A0-\u20CF',
    'Combining Marks for Symbols': '\u20D0-\u20FF',
    'Letterlike Symbols': '\u2100-\u214F',
    'Number Forms': '\u2150-\u218F',
    'Arrows': '\u2190-\u21FF',
    'Mathematical Operators': '\u2200-\u22FF',
    'Miscellaneous Technical': '\u2300-\u23FF',
    'Control Pictures': '\u2400-\u243F',
    'Optical Character Recognition': '\u2440-\u245F',
    'Enclosed Alphanumerics': '\u2460-\u24FF',
    'Box Drawing': '\u2500-\u257F',
    'Block Elements': '\u2580-\u259F',
    'Geometric Shapes': '\u25A0-\u25FF',
    'Miscellaneous Symbols': '\u2600-\u26FF',
    'Dingbats': '\u2700-\u27BF',
    'CJK Symbols and Punctuation': '\u3000-\u303F',
    'Hiragana': '\u3040-\u309F',
    'Katakana': '\u30A0-\u30FF',
    'Bopomofo': '\u3100-\u312F',
    'Hangul Compatibility Jamo': '\u3130-\u318F',
    'Kanbun': '\u3190-\u319F',
    'Enclosed CJK Letters and Months': '\u3200-\u32FF',
    'CJK Compatibility': '\u3300-\u33FF',
    'CJK Unified Ideographs': '\u4E00-\u9FFF',
    'Hangul Syllables': '\uAC00-\uD7A3',
    'High Surrogates': '\uD800-\uDB7F',
    'High Private Use Surrogates': '\uDB80-\uDBFF',
    'Low Surrogates': '\uDC00-\uDFFF',
    'Private Use': '\uE000-\uF8FF',
    'CJK Compatibility Ideographs': '\uF900-\uFAFF',
    'Alphabetic Presentation Forms': '\uFB00-\uFB4F',
    'Arabic Presentation Forms-A': '\uFB50-\uFDFF',
    'Combining Half Marks': '\uFE20-\uFE2F',
    'CJK Compatibility Forms': '\uFE30-\uFE4F',
    'Small Form Variants': '\uFE50-\uFE6F',
    'Arabic Presentation Forms-B': '\uFE70-\uFEFF',
    'Halfwidth and Fullwidth Forms': '\uFF00-\uFFEF',
    'Specials': '\uFEFF-\uFEFF\uFFF0-\uFFFF'
}

UPDATE_BLOCKS_VER_2_1_9 = {
    'Arabic Presentation Forms-B': '\uFE70-\uFEFE',
    'Specials': '\uFEFF-\uFEFF\uFFF0-\uFFFD'
}

UPDATE_BLOCKS_VER_3_0_0 = {
    'Syriac': '\u0700-\u074F',
    'Thaana': '\u0780-\u07BF',
    'Sinhala': '\u0D80-\u0DFF',
    'Tibetan': '\u0F00-\u0FFF',
    'Myanmar': '\u1000-\u109F',
    'Ethiopic': '\u1200-\u137F',
    'Cherokee': '\u13A0-\u13FF',
    'Unified Canadian Aboriginal Syllabics': '\u1400-\u167F',
    'Ogham': '\u1680-\u169F',
    'Runic': '\u16A0-\u16FF',
    'Khmer': '\u1780-\u17FF',
    'Mongolian': '\u1800-\u18AF',
    'Braille Patterns': '\u2800-\u28FF',
    'CJK Radicals Supplement': '\u2E80-\u2EFF',
    'Kangxi Radicals': '\u2F00-\u2FDF',
    'Ideographic Description Characters': '\u2FF0-\u2FFF',
    'Bopomofo Extended': '\u31A0-\u31BF',
    'CJK Unified Ideographs Extension A': '\u3400-\u4DB5',
    'Yi Syllables': '\uA000-\uA48F',
    'Yi Radicals': '\uA490-\uA4CF'
}

UPDATE_BLOCKS_VER_3_1_0 = {
    'Private Use': '\uE000-\uF8FF\U000F0000-\U000FFFFD\U00100000-\U0010FFFD',
    'Old Italic': '\U00010300-\U0001032F',
    'Gothic': '\U00010330-\U0001034F',
    'Deseret': '\U00010400-\U0001044F',
    'Byzantine Musical Symbols': '\U0001D000-\U0001D0FF',
    'Musical Symbols': '\U0001D100-\U0001D1FF',
    'Mathematical Alphanumeric Symbols': '\U0001D400-\U0001D7FF',
    'CJK Unified Ideographs Extension B': '\U00020000-\U0002A6D6',
    'CJK Compatibility Ideographs Supplement': '\U0002F800-\U0002FA1F',
    'Tags': '\U000E0000-\U000E007F'
}

REMOVED_BLOCKS_VER_3_1_1 = [
    'Greek', 'Combining Marks for Symbols', 'Private Use'
]

UPDATE_BLOCKS_VER_3_1_1 = {
    'Greek and Coptic': '\u0370-\u03FF',
    'Cyrillic Supplementary': '\u0500-\u052F',
    'Tagalog': '\u1700-\u171F',
    'Hanunoo': '\u1720-\u173F',
    'Buhid': '\u1740-\u175F',
    'Tagbanwa': '\u1760-\u177F',
    'Combining Diacritical Marks for Symbols': '\u20D0-\u20FF',
    'Miscellaneous Mathematical Symbols-A': '\u27C0-\u27EF',
    'Supplemental Arrows-A': '\u27F0-\u27FF',
    'Supplemental Arrows-B': '\u2900-\u297F',
    'Miscellaneous Mathematical Symbols-B': '\u2980-\u29FF',
    'Supplemental Mathematical Operators': '\u2A00-\u2AFF',
    'Katakana Phonetic Extensions': '\u31F0-\u31FF',
    'CJK Unified Ideographs Extension A': '\u3400-\u4DBF',
    'Hangul Syllables': '\uAC00-\uD7AF',
    'Private Use Area': '\uE000-\uF8FF',
    'Variation Selectors': '\uFE00-\uFE0F',
    'Arabic Presentation Forms-B': '\uFE70-\uFEFF',
    'Specials': '\uFFF0-\uFFFF',
    'CJK Unified Ideographs Extension B': '\U00020000-\U0002A6DF',
    'Supplementary Private Use Area-A': '\U000F0000-\U000FFFFF',
    'Supplementary Private Use Area-B': '\U00100000-\U0010FFFF'
}

UPDATE_BLOCKS_VER_4_0_0 = {
    'Limbu': '\u1900-\u194F',
    'Tai Le': '\u1950-\u197F',
    'Khmer Symbols': '\u19E0-\u19FF',
    'Phonetic Extensions': '\u1D00-\u1D7F',
    'Miscellaneous Symbols and Arrows': '\u2B00-\u2BFF',
    'Yijing Hexagram Symbols': '\u4DC0-\u4DFF',
    'Linear B Syllabary': '\U00010000-\U0001007F',
    'Linear B Ideograms': '\U00010080-\U000100FF',
    'Aegean Numbers': '\U00010100-\U0001013F',
    'Ugaritic': '\U00010380-\U0001039F',
    'Shavian': '\U00010450-\U0001047F',
    'Osmanya': '\U00010480-\U000104AF',
    'Cypriot Syllabary': '\U00010800-\U0001083F',
    'Tai Xuan Jing Symbols': '\U0001D300-\U0001D35F',
    'Variation Selectors Supplement': '\U000E0100-\U000E01EF'
}

REMOVED_BLOCKS_VER_4_0_1 = [
    'Cyrillic Supplementary'
]

UPDATE_BLOCKS_VER_4_0_1 = {
    'Cyrillic Supplement': '\u0500-\u052F'
}

UPDATE_BLOCKS_VER_4_1_0 = {
    'Arabic Supplement': '\u0750-\u077F',
    'Ethiopic Supplement': '\u1380-\u139F',
    'New Tai Lue': '\u1980-\u19DF',
    'Buginese': '\u1A00-\u1A1F',
    'Phonetic Extensions Supplement': '\u1D80-\u1DBF',
    'Combining Diacritical Marks Supplement': '\u1DC0-\u1DFF',
    'Glagolitic': '\u2C00-\u2C5F',
    'Coptic': '\u2C80-\u2CFF',
    'Georgian Supplement': '\u2D00-\u2D2F',
    'Tifinagh': '\u2D30-\u2D7F',
    'Ethiopic Extended': '\u2D80-\u2DDF',
    'Supplemental Punctuation': '\u2E00-\u2E7F',
    'CJK Strokes': '\u31C0-\u31EF',
    'Modifier Tone Letters': '\uA700-\uA71F',
    'Syloti Nagri': '\uA800-\uA82F',
    'Vertical Forms': '\uFE10-\uFE1F',
    'Ancient Greek Numbers': '\U00010140-\U0001018F',
    'Old Persian': '\U000103A0-\U000103DF',
    'Kharoshthi': '\U00010A00-\U00010A5F',
    'Ancient Greek Musical Notation': '\U0001D200-\U0001D24F'
}

UPDATE_BLOCKS_VER_5_0_0 = {
    'NKo': '\u07C0-\u07FF',
    'Balinese': '\u1B00-\u1B7F',
    'Latin Extended-C': '\u2C60-\u2C7F',
    'Latin Extended-D': '\uA720-\uA7FF',
    'Phags-pa': '\uA840-\uA87F',
    'Phoenician': '\U00010900-\U0001091F',
    'Cuneiform': '\U00012000-\U000123FF',
    'Cuneiform Numbers and Punctuation': '\U00012400-\U0001247F',
    'Counting Rod Numerals': '\U0001D360-\U0001D37F'
}

UPDATE_BLOCKS_VER_5_1_0 = {
    'Sundanese': '\u1B80-\u1BBF',
    'Lepcha': '\u1C00-\u1C4F',
    'Ol Chiki': '\u1C50-\u1C7F',
    'Cyrillic Extended-A': '\u2DE0-\u2DFF',
    'Vai': '\uA500-\uA63F',
    'Cyrillic Extended-B': '\uA640-\uA69F',
    'Saurashtra': '\uA880-\uA8DF',
    'Kayah Li': '\uA900-\uA92F',
    'Rejang': '\uA930-\uA95F',
    'Cham': '\uAA00-\uAA5F',
    'Ancient Symbols': '\U00010190-\U000101CF',
    'Phaistos Disc': '\U000101D0-\U000101FF',
    'Lycian': '\U00010280-\U0001029F',
    'Carian': '\U000102A0-\U000102DF',
    'Lydian': '\U00010920-\U0001093F',
    'Mahjong Tiles': '\U0001F000-\U0001F02F',
    'Domino Tiles': '\U0001F030-\U0001F09F'
}

UPDATE_BLOCKS_VER_5_2_0 = {
    'Samaritan': '\u0800-\u083F',
    'Unified Canadian Aboriginal Syllabics Extended': '\u18B0-\u18FF',
    'Tai Tham': '\u1A20-\u1AAF',
    'Vedic Extensions': '\u1CD0-\u1CFF',
    'Lisu': '\uA4D0-\uA4FF',
    'Bamum': '\uA6A0-\uA6FF',
    'Common Indic Number Forms': '\uA830-\uA83F',
    'Devanagari Extended': '\uA8E0-\uA8FF',
    'Hangul Jamo Extended-A': '\uA960-\uA97F',
    'Javanese': '\uA980-\uA9DF',
    'Myanmar Extended-A': '\uAA60-\uAA7F',
    'Tai Viet': '\uAA80-\uAADF',
    'Meetei Mayek': '\uABC0-\uABFF',
    'Hangul Jamo Extended-B': '\uD7B0-\uD7FF',
    'Imperial Aramaic': '\U00010840-\U0001085F',
    'Old South Arabian': '\U00010A60-\U00010A7F',
    'Avestan': '\U00010B00-\U00010B3F',
    'Inscriptional Parthian': '\U00010B40-\U00010B5F',
    'Inscriptional Pahlavi': '\U00010B60-\U00010B7F',
    'Old Turkic': '\U00010C00-\U00010C4F',
    'Rumi Numeral Symbols': '\U00010E60-\U00010E7F',
    'Kaithi': '\U00011080-\U000110CF',
    'Egyptian Hieroglyphs': '\U00013000-\U0001342F',
    'Enclosed Alphanumeric Supplement': '\U0001F100-\U0001F1FF',
    'Enclosed Ideographic Supplement': '\U0001F200-\U0001F2FF',
    'CJK Unified Ideographs Extension C': '\U0002A700-\U0002B73F'
}

UPDATE_BLOCKS_VER_6_0_0 = {
    'Mandaic': '\u0840-\u085F',
    'Batak': '\u1BC0-\u1BFF',
    'Ethiopic Extended-A': '\uAB00-\uAB2F',
    'Brahmi': '\U00011000-\U0001107F',
    'Bamum Supplement': '\U00016800-\U00016A3F',
    'Kana Supplement': '\U0001B000-\U0001B0FF',
    'Playing Cards': '\U0001F0A0-\U0001F0FF',
    'Miscellaneous Symbols And Pictographs': '\U0001F300-\U0001F5FF',
    'Emoticons': '\U0001F600-\U0001F64F',
    'Transport And Map Symbols': '\U0001F680-\U0001F6FF',
    'Alchemical Symbols': '\U0001F700-\U0001F77F',
    'CJK Unified Ideographs Extension D': '\U0002B740-\U0002B81F'
}

UPDATE_BLOCKS_VER_6_1_0 = {
    'Arabic Extended-A': '\u08A0-\u08FF',
    'Sundanese Supplement': '\u1CC0-\u1CCF',
    'Meetei Mayek Extensions': '\uAAE0-\uAAFF',
    'Meroitic Hieroglyphs': '\U00010980-\U0001099F',
    'Meroitic Cursive': '\U000109A0-\U000109FF',
    'Sora Sompeng': '\U000110D0-\U000110FF',
    'Chakma': '\U00011100-\U0001114F',
    'Sharada': '\U00011180-\U000111DF',
    'Takri': '\U00011680-\U000116CF',
    'Miao': '\U00016F00-\U00016F9F',
    'Arabic Mathematical Alphabetic Symbols': '\U0001EE00-\U0001EEFF'
}

REMOVED_BLOCKS_VER_7_0_0 = [
    'Miscellaneous Symbols And Pictographs', 'Transport And Map Symbols'
]

UPDATE_BLOCKS_VER_7_0_0 = {
    'Combining Diacritical Marks Extended': '\u1AB0-\u1AFF',
    'Myanmar Extended-B': '\uA9E0-\uA9FF',
    'Latin Extended-E': '\uAB30-\uAB6F',
    'Coptic Epact Numbers': '\U000102E0-\U000102FF',
    'Old Permic': '\U00010350-\U0001037F',
    'Elbasan': '\U00010500-\U0001052F',
    'Caucasian Albanian': '\U00010530-\U0001056F',
    'Linear A': '\U00010600-\U0001077F',
    'Palmyrene': '\U00010860-\U0001087F',
    'Nabataean': '\U00010880-\U000108AF',
    'Old North Arabian': '\U00010A80-\U00010A9F',
    'Manichaean': '\U00010AC0-\U00010AFF',
    'Psalter Pahlavi': '\U00010B80-\U00010BAF',
    'Mahajani': '\U00011150-\U0001117F',
    'Sinhala Archaic Numbers': '\U000111E0-\U000111FF',
    'Khojki': '\U00011200-\U0001124F',
    'Khudawadi': '\U000112B0-\U000112FF',
    'Grantha': '\U00011300-\U0001137F',
    'Tirhuta': '\U00011480-\U000114DF',
    'Siddham': '\U00011580-\U000115FF',
    'Modi': '\U00011600-\U0001165F',
    'Warang Citi': '\U000118A0-\U000118FF',
    'Pau Cin Hau': '\U00011AC0-\U00011AFF',
    'Mro': '\U00016A40-\U00016A6F',
    'Bassa Vah': '\U00016AD0-\U00016AFF',
    'Pahawh Hmong': '\U00016B00-\U00016B8F',
    'Duployan': '\U0001BC00-\U0001BC9F',
    'Shorthand Format Controls': '\U0001BCA0-\U0001BCAF',
    'Mende Kikakui': '\U0001E800-\U0001E8DF',
    'Miscellaneous Symbols and Pictographs': '\U0001F300-\U0001F5FF',
    'Ornamental Dingbats': '\U0001F650-\U0001F67F',
    'Transport and Map Symbols': '\U0001F680-\U0001F6FF',
    'Geometric Shapes Extended': '\U0001F780-\U0001F7FF',
    'Supplemental Arrows-C': '\U0001F800-\U0001F8FF'
}

UPDATE_BLOCKS_VER_8_0_0 = {
    'Cherokee Supplement': '\uAB70-\uABBF',
    'Hatran': '\U000108E0-\U000108FF',
    'Old Hungarian': '\U00010C80-\U00010CFF',
    'Multani': '\U00011280-\U000112AF',
    'Ahom': '\U00011700-\U0001173F',
    'Early Dynastic Cuneiform': '\U00012480-\U0001254F',
    'Anatolian Hieroglyphs': '\U00014400-\U0001467F',
    'Sutton SignWriting': '\U0001D800-\U0001DAAF',
    'Supplemental Symbols and Pictographs': '\U0001F900-\U0001F9FF',
    'CJK Unified Ideographs Extension E': '\U0002B820-\U0002CEAF'
}

UPDATE_BLOCKS_VER_9_0_0 = {
    'Cyrillic Extended-C': '\u1C80-\u1C8F',
    'Osage': '\U000104B0-\U000104FF',
    'Newa': '\U00011400-\U0001147F',
    'Mongolian Supplement': '\U00011660-\U0001167F',
    'Bhaiksuki': '\U00011C00-\U00011C6F',
    'Marchen': '\U00011C70-\U00011CBF',
    'Ideographic Symbols and Punctuation': '\U00016FE0-\U00016FFF',
    'Tangut': '\U00017000-\U000187FF',
    'Tangut Components': '\U00018800-\U00018AFF',
    'Glagolitic Supplement': '\U0001E000-\U0001E02F',
    'Adlam': '\U0001E900-\U0001E95F'
}

UPDATE_BLOCKS_VER_10_0_0 = {
    'Syriac Supplement': '\u0860-\u086F',
    'Zanabazar Square': '\U00011A00-\U00011A4F',
    'Soyombo': '\U00011A50-\U00011AAF',
    'Masaram Gondi': '\U00011D00-\U00011D5F',
    'Kana Extended-A': '\U0001B100-\U0001B12F',
    'Nushu': '\U0001B170-\U0001B2FF',
    'CJK Unified Ideographs Extension F': '\U0002CEB0-\U0002EBEF'
}

UPDATE_BLOCKS_VER_11_0_0 = {
    'Georgian Extended': '\u1C90-\u1CBF',
    'Hanifi Rohingya': '\U00010D00-\U00010D3F',
    'Old Sogdian': '\U00010F00-\U00010F2F',
    'Sogdian': '\U00010F30-\U00010F6F',
    'Dogra': '\U00011800-\U0001184F',
    'Gunjala Gondi': '\U00011D60-\U00011DAF',
    'Makasar': '\U00011EE0-\U00011EFF',
    'Medefaidrin': '\U00016E40-\U00016E9F',
    'Mayan Numerals': '\U0001D2E0-\U0001D2FF',
    'Indic Siyaq Numbers': '\U0001EC70-\U0001ECBF',
    'Chess Symbols': '\U0001FA00-\U0001FA6F'
}

UPDATE_BLOCKS_VER_12_0_0 = {
    'Elymaic': '\U00010FE0-\U00010FFF',
    'Nandinagari': '\U000119A0-\U000119FF',
    'Tamil Supplement': '\U00011FC0-\U00011FFF',
    'Egyptian Hieroglyph Format Controls': '\U00013430-\U0001343F',
    'Small Kana Extension': '\U0001B130-\U0001B16F',
    'Nyiakeng Puachue Hmong': '\U0001E100-\U0001E14F',
    'Wancho': '\U0001E2C0-\U0001E2FF',
    'Ottoman Siyaq Numbers': '\U0001ED00-\U0001ED4F',
    'Symbols and Pictographs Extended-A': '\U0001FA70-\U0001FAFF'
}

UPDATE_BLOCKS_VER_13_0_0 = {
    'Yezidi': '\U00010E80-\U00010EBF',
    'Chorasmian': '\U00010FB0-\U00010FDF',
    'Dives Akuru': '\U00011900-\U0001195F',
    'Lisu Supplement': '\U00011FB0-\U00011FBF',
    'Khitan Small Script': '\U00018B00-\U00018CFF',
    'Tangut Supplement': '\U00018D00-\U00018D8F',
    'Symbols for Legacy Computing': '\U0001FB00-\U0001FBFF',
    'CJK Unified Ideographs Extension G': '\U00030000-\U0003134F'
}

UPDATE_BLOCKS_VER_14_0_0 = {
    'Arabic Extended-B': '\u0870-\u089F',
    'Vithkuqi': '\U00010570-\U000105BF',
    'Latin Extended-F': '\U00010780-\U000107BF',
    'Old Uyghur': '\U00010F70-\U00010FAF',
    'Ahom': '\U00011700-\U0001174F',
    'Unified Canadian Aboriginal Syllabics Extended-A': '\U00011AB0-\U00011ABF',
    'Cypro-Minoan': '\U00012F90-\U00012FFF',
    'Tangsa': '\U00016A70-\U00016ACF',
    'Tangut Supplement': '\U00018D00-\U00018D7F',
    'Kana Extended-B': '\U0001AFF0-\U0001AFFF',
    'Znamenny Musical Notation': '\U0001CF00-\U0001CFCF',
    'Latin Extended-G': '\U0001DF00-\U0001DFFF',
    'Toto': '\U0001E290-\U0001E2BF',
    'Ethiopic Extended-B': '\U0001E7E0-\U0001E7FF'
}

UPDATE_BLOCKS_VER_15_0_0 = {
    'Arabic Extended-C': '\U00010EC0-\U00010EFF',
    'Devanagari Extended-A': '\U00011B00-\U00011B5F',
    'Kawi': '\U00011F00-\U00011F5F',
    'Egyptian Hieroglyph Format Controls': '\U00013430-\U0001345F',
    'Kaktovik Numerals': '\U0001D2C0-\U0001D2DF',
    'Cyrillic Extended-D': '\U0001E030-\U0001E08F',
    'Nag Mundari': '\U0001E4D0-\U0001E4FF',
    'CJK Unified Ideographs Extension H': '\U00031350-\U000323AF'
}

UPDATE_BLOCKS_VER_15_1_0 = {
    'CJK Unified Ideographs Extension I': '\U0002EBF0-\U0002EE5F'
}

UPDATE_BLOCKS_VER_16_0_0 = {
    'Todhri': '\U000105C0-\U000105FF',
    'Garay': '\U00010D40-\U00010D8F',
    'Tulu-Tigalari': '\U00011380-\U000113FF',
    'Myanmar Extended-C': '\U000116D0-\U000116FF',
    'Sunuwar': '\U00011BC0-\U00011BFF',
    'Egyptian Hieroglyphs Extended-A': '\U00013460-\U000143FF',
    'Gurung Khema': '\U00016100-\U0001613F',
    'Kirat Rai': '\U00016D40-\U00016D7F',
    'Symbols for Legacy Computing Supplement': '\U0001CC00-\U0001CEBF',
    'Ol Onal': '\U0001E5D0-\U0001E5FF'
}

UPDATE_BLOCKS_VER_17_0_0 = {
    'Sidetic': '\U00010940-\U0001095F',
    'Sharada Supplement': '\U00011B60-\U00011B7F',
    'Tolong Siki': '\U00011DB0-\U00011DEF',
    'Beria Erfe': '\U00016EA0-\U00016EDF',
    'Tangut Components Supplement': '\U00018D80-\U00018DFF',
    'Miscellaneous Symbols Supplement': '\U0001CEC0-\U0001CEFF',
    'Tai Yo': '\U0001E6C0-\U0001E6FF',
    'CJK Unified Ideographs Extension J': '\U000323B0-\U0003347F'
}



#
# Copyright (c), 2018-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
from urllib.parse import urlparse
from typing import Union

from elementpath.helpers import collapse_white_spaces, Patterns
from .any_types import AnyAtomicType
from .untyped import UntypedAtomic

__all__ = ['AnyURI']


class AnyURI(AnyAtomicType):
    value: str
    name = 'anyURI'

    __slots__ = ('value',)

    def __init__(self, value: Union[str, bytes, UntypedAtomic, 'AnyURI']) -> None:
        """
        :param value: string or an untyped atomic that represents a valid URI.
        """
        match value:
            case str():
                self.value = collapse_white_spaces(value)
            case bytes():
                self.value = collapse_white_spaces(value.decode('utf-8'))
            case UntypedAtomic():
                self.value = collapse_white_spaces(value.value)
            case AnyURI():
                self.value = value.value
            case _:
                raise TypeError('the argument has an invalid type %r' % type(value))

        self.validate(self.value)

    def __repr__(self) -> str:
        return '%s(%r)' % (self.__class__.__name__, self.value)

    def __str__(self) -> str:
        return self.value

    def __bool__(self) -> bool:
        return bool(self.value)  # For effective boolean value

    def __hash__(self) -> int:
        return hash(self.value)

    def __contains__(self, item: str) -> bool:
        return item in self.value

    def __eq__(self, other: object) -> bool:
        if isinstance(other, (AnyURI, UntypedAtomic)):
            return self.value == other.value
        elif isinstance(other, str):
            return self.value == other
        return NotImplemented

    def __ne__(self, other: object) -> bool:
        if isinstance(other, (AnyURI, UntypedAtomic)):
            return self.value != other.value
        elif isinstance(other, str):
            return self.value != other
        return NotImplemented

    def __lt__(self, other: Union[str, 'AnyURI', UntypedAtomic]) -> bool:
        if isinstance(other, (AnyURI, UntypedAtomic)):
            return self.value < other.value
        return self.value < other

    def __le__(self, other: Union[str, 'AnyURI', UntypedAtomic]) -> bool:
        if isinstance(other, (AnyURI, UntypedAtomic)):
            return self.value <= other.value
        return self.value <= other

    def __gt__(self, other: Union[str, 'AnyURI', UntypedAtomic]) -> bool:
        if isinstance(other, (AnyURI, UntypedAtomic)):
            return self.value > other.value
        return self.value > other

    def __ge__(self, other: Union[str, 'AnyURI', UntypedAtomic]) -> bool:
        if isinstance(other, (AnyURI, UntypedAtomic)):
            return self.value >= other.value
        return self.value >= other

    @classmethod
    def validate(cls, value: object) -> None:
        if isinstance(value, cls):
            return
        elif isinstance(value, bytes):
            value = value.decode()
        elif not isinstance(value, str):
            raise cls._invalid_type(value)

        try:
            url_parts = urlparse(value)
            _ = url_parts.port  # check invalid port!
        except ValueError as err:
            msg = 'invalid value {!r} for xs:{} ({})'
            raise ValueError(msg.format(value, cls.name, str(err))) from None
        else:
            if url_parts.path.startswith(':'):
                raise cls._invalid_value(value)
            elif value.count('#') > 1:
                msg = 'invalid value {!r} for xs:{} (too many # characters)'
                raise ValueError(msg.format(value, cls.name))
            elif Patterns.wrong_escape.search(value) is not None:
                msg = 'invalid value {!r} for xs:{} (wrong escaping)'
                raise ValueError(msg.format(value, cls.name))



#
# Copyright (c), 2018-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
"""
Subpackage for validating against XPath standard schemas.
"""
import pathlib
from xml.etree.ElementTree import Element
from typing import Optional

from elementpath.exceptions import ElementPathRuntimeError

try:
    import xmlschema
except ImportError:  # pragma: no cover
    from ..exceptions import xpath_error

    def validate_analyzed_string(root: Element) -> None:
        raise ElementPathRuntimeError('not schema-aware')

    def validate_json_to_xml(root: Element) -> None:
        raise xpath_error('FOJS0004')

else:
    from ..namespaces import XPATH_FUNCTIONS_NAMESPACE

    analyzed_string_schema: Optional[xmlschema.XMLSchemaBase] = None
    json_to_xml_schema: Optional[xmlschema.XMLSchemaBase] = None

    __all__ = ['validate_analyzed_string', 'validate_json_to_xml']

    def validate_analyzed_string(root: Element) -> None:
        global analyzed_string_schema

        if analyzed_string_schema is None:
            xsd_file = pathlib.Path(__file__).parent.joinpath('analyze-string.xsd')
            analyzed_string_schema = xmlschema.XMLSchema(xsd_file)

        analyzed_string_schema.validate(root)

    def validate_json_to_xml(root: Element) -> None:
        global json_to_xml_schema

        if json_to_xml_schema is None:
            xsd_file = pathlib.Path(__file__).parent.joinpath('schema-for-json.xsd')
            json_to_xml_schema = xmlschema.XMLSchema(xsd_file)

        json_to_xml_schema.validate(root, namespaces={'j': XPATH_FUNCTIONS_NAMESPACE})



#
# Copyright (c), 2024-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
"""
A module for providing XSD atomic values for schema context evaluation
and for decoding text into XSD atomic data values.
"""
from collections.abc import Iterator
from decimal import Decimal
from typing import Optional, Union

from elementpath import aliases
from elementpath.aliases import AnyNsmapType
from elementpath.protocols import XsdTypeProtocol
from elementpath.exceptions import xpath_error
from elementpath.namespaces import XSD_NAMESPACE
import elementpath.datatypes as dt


class _Notation(dt.Notation):
    """An instantiable xs:NOTATION."""


_ATOMIC_VALUES: dict[str, dict[str, aliases.AtomicType]] = {'1.0': {
    f'{{{XSD_NAMESPACE}}}untypedAtomic': dt.UntypedAtomic('1'),
    f'{{{XSD_NAMESPACE}}}anyType': dt.UntypedAtomic('1'),
    f'{{{XSD_NAMESPACE}}}anySimpleType': dt.UntypedAtomic('1'),
    f'{{{XSD_NAMESPACE}}}anyAtomicType': dt.UntypedAtomic('1'),
    f'{{{XSD_NAMESPACE}}}boolean': True,
    f'{{{XSD_NAMESPACE}}}decimal': Decimal('1.0'),
    f'{{{XSD_NAMESPACE}}}double': float('1.0'),
    f'{{{XSD_NAMESPACE}}}float': dt.Float('1.0'),
    f'{{{XSD_NAMESPACE}}}string': '  alpha\t',
    f'{{{XSD_NAMESPACE}}}date': dt.Date10.fromstring('2000-01-01'),
    f'{{{XSD_NAMESPACE}}}dateTime': dt.DateTime10.fromstring('2000-01-01T12:00:00'),
    f'{{{XSD_NAMESPACE}}}gDay': dt.GregorianDay.fromstring('---31'),
    f'{{{XSD_NAMESPACE}}}gMonth': dt.GregorianMonth.fromstring('--12'),
    f'{{{XSD_NAMESPACE}}}gMonthDay': dt.GregorianMonthDay.fromstring('--12-01'),
    f'{{{XSD_NAMESPACE}}}gYear': dt.GregorianYear10.fromstring('1999'),
    f'{{{XSD_NAMESPACE}}}gYearMonth': dt.GregorianYearMonth10.fromstring('1999-09'),
    f'{{{XSD_NAMESPACE}}}time': dt.Time.fromstring('09:26:54'),
    f'{{{XSD_NAMESPACE}}}duration': dt.Duration.fromstring('P1MT1S'),
    f'{{{XSD_NAMESPACE}}}dayTimeDuration': dt.DayTimeDuration.fromstring('P1DT1S'),
    f'{{{XSD_NAMESPACE}}}yearMonthDuration': dt.YearMonthDuration.fromstring('P1Y1M'),
    f'{{{XSD_NAMESPACE}}}QName': dt.QName(XSD_NAMESPACE, 'xs:element'),
    f'{{{XSD_NAMESPACE}}}NOTATION': _Notation(XSD_NAMESPACE, 'xs:element'),
    f'{{{XSD_NAMESPACE}}}anyURI': dt.AnyURI('https://example.com'),
    f'{{{XSD_NAMESPACE}}}normalizedString': dt.NormalizedString(' alpha  '),
    f'{{{XSD_NAMESPACE}}}token': dt.XsdToken('a token'),
    f'{{{XSD_NAMESPACE}}}language': dt.Language('en-US'),
    f'{{{XSD_NAMESPACE}}}Name': dt.Name('_a.name::'),
    f'{{{XSD_NAMESPACE}}}NCName': dt.NCName('nc-name'),
    f'{{{XSD_NAMESPACE}}}ID': dt.Id('id1'),
    f'{{{XSD_NAMESPACE}}}IDREF': dt.Idref('id_ref1'),
    f'{{{XSD_NAMESPACE}}}ENTITY': dt.Entity('entity1'),
    f'{{{XSD_NAMESPACE}}}NMTOKEN': dt.NMToken('a_token'),
    f'{{{XSD_NAMESPACE}}}base64Binary': dt.Base64Binary('YWxwaGE='),
    f'{{{XSD_NAMESPACE}}}hexBinary': dt.HexBinary('31'),
    f'{{{XSD_NAMESPACE}}}integer': dt.Integer('1'),
    f'{{{XSD_NAMESPACE}}}long': dt.Long('1'),
    f'{{{XSD_NAMESPACE}}}int': dt.Int('1'),
    f'{{{XSD_NAMESPACE}}}short': dt.Short('1'),
    f'{{{XSD_NAMESPACE}}}byte': dt.Byte('1'),
    f'{{{XSD_NAMESPACE}}}positiveInteger': dt.PositiveInteger('1'),
    f'{{{XSD_NAMESPACE}}}negativeInteger': dt.NegativeInteger('-1'),
    f'{{{XSD_NAMESPACE}}}nonPositiveInteger': dt.NonPositiveInteger('0'),
    f'{{{XSD_NAMESPACE}}}nonNegativeInteger': dt.NonNegativeInteger('0'),
    f'{{{XSD_NAMESPACE}}}unsignedLong': dt.UnsignedLong('1'),
    f'{{{XSD_NAMESPACE}}}unsignedInt': dt.UnsignedInt('1'),
    f'{{{XSD_NAMESPACE}}}unsignedShort': dt.UnsignedShort('1'),
    f'{{{XSD_NAMESPACE}}}unsignedByte': dt.UnsignedByte('1'),
}}

_ATOMIC_VALUES['1.1'] = {
    **_ATOMIC_VALUES['1.0'],
    f'{{{XSD_NAMESPACE}}}date': dt.Date.fromstring('2000-01-01'),
    f'{{{XSD_NAMESPACE}}}dateTime': dt.DateTime.fromstring('2000-01-01T12:00:00'),
    f'{{{XSD_NAMESPACE}}}gYear': dt.GregorianYear.fromstring('1999'),
    f'{{{XSD_NAMESPACE}}}gYearMonth': dt.GregorianYearMonth.fromstring('1999-09'),
    f'{{{XSD_NAMESPACE}}}dateTimeStamp': dt.DateTimeStamp.fromstring('2000-01-01T12:00:00+01:00'),
}

# List based types
_LIST_VALUES = {
    f'{{{XSD_NAMESPACE}}}IDREFS': dt.Idrefs([dt.Idref('id_ref1'), dt.Idref('id_ref2')]),
    f'{{{XSD_NAMESPACE}}}ENTITIES': dt.Entities([dt.Entity('entity1'), dt.Entity('entity2')]),
    f'{{{XSD_NAMESPACE}}}NMTOKENS': dt.NMTokens([dt.NMToken('a_token'), dt.NMToken('b_token')]),
}


def iter_atomic_values(xsd_type: XsdTypeProtocol) -> Iterator[aliases.AtomicType]:
    """Generates a list of XSD atomic values related to provided XSD type."""

    def _iter_values(root_type: XsdTypeProtocol, depth: int) -> Iterator[aliases.AtomicType]:
        if depth > 15:
            return
        if root_type.name in atomic_values:
            yield atomic_values[root_type.name]
        elif hasattr(root_type, 'member_types'):
            for member_type in root_type.member_types:
                yield from _iter_values(member_type, depth + 1)

    atomic_values = _ATOMIC_VALUES[xsd_type.xsd_version]
    if xsd_type.name in atomic_values:
        yield atomic_values[xsd_type.name]
    elif xsd_type.is_simple() or (simple_type := xsd_type.simple_type) is None:
        yield from _iter_values(xsd_type.root_type, 1)
    elif simple_type.name in atomic_values:
        yield atomic_values[simple_type.name]
    else:
        yield from _iter_values(simple_type.root_type, 1)


def get_atomic_sequence(xsd_type: Optional[XsdTypeProtocol],
                        text: Optional[str] = None,
                        namespaces: AnyNsmapType = None) -> Iterator[aliases.AtomicType]:
    """Returns a decoder function for atomic values of an XSD type instance."""
    def decode(s: str) -> aliases.AtomicType:
        if isinstance(value, (dt.AbstractDateTime, dt.Duration)):
            return value.fromstring(s)
        elif not isinstance(value, dt.AbstractQName):
            return value.__class__(s)
        else:
            nonlocal namespaces
            if namespaces is None:
                namespaces = {}
            if ':' not in s:
                return value.__class__(namespaces.get(''), s)
            else:
                return value.__class__(namespaces[s.split(':')[0]], s)

    if xsd_type is None:
        yield dt.UntypedAtomic(text or '')
    elif text is None:
        yield from iter_atomic_values(xsd_type)
    else:
        error: Union[None, ValueError, ArithmeticError] = None
        code = 'FORG0001'

        for value in iter_atomic_values(xsd_type):
            try:
                if xsd_type.is_list():
                    for item in text.split():
                        yield decode(item)
                else:
                    yield decode(text)
            except (ArithmeticError, ValueError) as err:
                if error is None:
                    error = err
                    if isinstance(err, ArithmeticError):
                        if isinstance(value, dt.AbstractDateTime):
                            code = 'FODT0001'
                        elif isinstance(value, dt.Duration):
                            code = 'FODT0002'
                        else:
                            code = 'FOCA0002'
            else:
                return
        else:
            if error is not None:
                raise xpath_error(code, error, namespaces=namespaces)
            elif hasattr(xsd_type, 'decode'):
                yield xsd_type.decode(text or '')
            else:
                yield dt.UntypedAtomic(text if isinstance(text, str) else '')


__all__ = ['get_atomic_sequence']



#
# Copyright (c), 2018-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
"""
Type proxies for using builtin Python datatypes as primitive XSD datatypes.
A proxy class creates and validates a Python datatype and can be used for
checking instances of XSD primitive and derived types registered as virtual
subclasses.
"""
import math
from decimal import Decimal
from typing import Any, cast, NoReturn

from elementpath.aliases import XPath2ParserType
from elementpath.helpers import BOOLEAN_VALUES, FloatArgType, LazyPattern, \
    collapse_white_spaces, get_double
from .any_types import AnyAtomicType
from .untyped import UntypedAtomic
from .numeric import Float, Integer
from .datetime import AbstractDateTime, Duration

__all__ = ['BooleanProxy', 'DecimalProxy', 'DoubleProxy', 'DoubleProxy10',
           'StringProxy', 'NumericProxy', 'ArithmeticProxy', 'ErrorProxy']


class BooleanProxy(AnyAtomicType):
    """
    Proxy class for xs:boolean builtin datatype. Builds instances as <class 'bool'>.`.
    """
    name = 'boolean'
    pattern = LazyPattern(r'^(?:true|false|1|0)$')

    def __new__(cls, value: object) -> bool:  # type: ignore[misc]
        if isinstance(value, bool):
            return value
        elif isinstance(value, (int, float, Decimal)):
            if math.isnan(value):
                return False
            return bool(value)
        elif isinstance(value, UntypedAtomic):
            value = value.value
        elif not isinstance(value, str):
            raise TypeError('invalid type {!r} for xs:{}'.format(type(value), cls.name))

        if value.strip() not in BOOLEAN_VALUES:
            raise ValueError('invalid value {!r} for xs:{}'.format(value, cls.name))
        return 't' in value or '1' in value

    def __init__(self, value: object) -> None:
        bool.__init__(self)

    @classmethod
    def __subclasshook__(cls, subclass: type) -> bool:
        return issubclass(subclass, bool)

    @classmethod
    def validate(cls, value: object) -> None:
        if isinstance(value, bool):
            return
        elif isinstance(value, str):
            if cls.pattern.match(value) is None:
                raise cls._invalid_value(value)
        else:
            raise cls._invalid_type(value)


class DecimalProxy(AnyAtomicType):
    """
    Proxy class for xs:decimal builtin datatype. Builds instances as `decimal.Decimal`.
    """
    name = 'decimal'
    pattern = LazyPattern(r'^[+-]?(?:[0-9]+(?:\.[0-9]*)?|\.[0-9]+)$')

    def __new__(cls, value: Any) -> Decimal:  # type: ignore[misc]
        if isinstance(value, (str, UntypedAtomic)):
            value = collapse_white_spaces(str(value)).replace(' ', '')
            if cls.pattern.match(value) is None:
                raise cls._invalid_value(value)
        elif isinstance(value, (float, Float, Decimal)):
            if math.isinf(value) or math.isnan(value):
                raise cls._invalid_value(value)
        try:
            return Decimal(value)
        except (ValueError, ArithmeticError):
            msg = 'invalid value {!r} for xs:{}'
            raise ArithmeticError(msg.format(value, cls.name)) from None

    def __init__(self, value: Any) -> None:
        pass

    @classmethod
    def __subclasshook__(cls, subclass: type) -> bool:
        return issubclass(subclass, (int, Decimal, Integer)) and not issubclass(subclass, bool)

    @classmethod
    def validate(cls, value: object) -> None:
        match value:
            case Decimal():
                if math.isnan(value) or math.isinf(value):
                    raise cls._invalid_value(value)
            case bool():
                raise cls._invalid_type(value)
            case int() | Integer():
                return
            case str():
                if cls.pattern.match(value) is None:
                    raise cls._invalid_value(value)
            case _:
                raise cls._invalid_type(value)


class DoubleProxy(AnyAtomicType):
    """
    Proxy class for xs:double builtin datatype. Builds instances as <class 'float'>.`.
    """
    name = 'double'
    pattern = LazyPattern(
        r'^(?:[+-]?(?:[0-9]+(?:\.[0-9]*)?|\.[0-9]+)(?:[Ee][+-]?[0-9]+)?|[+-]?INF|NaN)$'
    )

    @classmethod
    def make(cls, value: Any,
             parser: XPath2ParserType | None = None,
             **kwargs: Any) -> 'float':
        if cls.__name__.endswith('10'):
            return get_double(value, '1.0')
        elif parser is not None:
            return get_double(value, parser.xsd_version)
        else:
            return get_double(value, kwargs.get('xsd_version'))

    def __new__(cls, value: FloatArgType) -> float:  # type: ignore[misc]
        return get_double(value)

    def __init__(self, value: FloatArgType) -> None:
        float.__init__(self)

    @classmethod
    def __subclasshook__(cls, subclass: type) -> bool:
        return issubclass(subclass, float) and not issubclass(subclass, Float)

    @classmethod
    def validate(cls, value: object) -> None:
        if isinstance(value, float) and not isinstance(value, Float):
            return
        elif isinstance(value, str):
            if cls.pattern.match(value) is None:
                raise cls._invalid_value(value)
        else:
            raise cls._invalid_type(value)


class DoubleProxy10(DoubleProxy):
    def __new__(cls, value: FloatArgType) -> float:  # type: ignore[misc]
        return get_double(value, '1.0')


class StringProxy(AnyAtomicType):
    """
    Proxy class for xs:string builtin datatype. Builds instances as <class 'str'>.`.
    """
    name = 'string'

    def __new__(cls, *args: object, **kwargs: object) -> str:  # type: ignore[misc]
        return str(*args, **kwargs)

    def __init__(self, *args: object, **kwargs: object) -> None:
        str.__init__(self)

    @classmethod
    def __subclasshook__(cls, subclass: type) -> bool:
        return issubclass(subclass, str)

    @classmethod
    def validate(cls, value: object) -> None:
        if not isinstance(value, str):
            raise cls._invalid_type(value)


####
# type proxies for multiple type-checking in XPath expressions
class NumericTypeMeta(type):
    """Metaclass for checking numeric classes and instances."""

    def __instancecheck__(cls, instance: object) -> bool:
        return isinstance(instance, (int, float, Decimal)) and not isinstance(instance, bool)

    def __subclasscheck__(cls, subclass: type) -> bool:
        if issubclass(subclass, bool):
            return False
        return issubclass(subclass, int) or issubclass(subclass, float) \
            or issubclass(subclass, Decimal)


class NumericProxy(metaclass=NumericTypeMeta):
    """Proxy for xs:numeric related types. Builds xs:float instances."""

    def __new__(cls, *args: FloatArgType, **kwargs: FloatArgType) -> float:  # type: ignore[misc]
        return float(*args, **kwargs)


class ArithmeticTypeMeta(type):
    """Metaclass for checking numeric, datetime and duration classes/instances."""

    def __instancecheck__(cls, instance: object) -> bool:
        return isinstance(
            instance, (int, float, Decimal, AbstractDateTime, Duration, UntypedAtomic)
        ) and not isinstance(instance, bool)

    def __subclasscheck__(cls, subclass: type) -> bool:
        if issubclass(subclass, bool):
            return False
        return issubclass(subclass, int) or issubclass(subclass, float) or \
            issubclass(subclass, Decimal) or issubclass(subclass, Duration) \
            or issubclass(subclass, AbstractDateTime) or issubclass(subclass, UntypedAtomic)


class ArithmeticProxy(metaclass=ArithmeticTypeMeta):
    """Proxy for arithmetic related types. Builds xs:float instances."""

    def __new__(cls, *args: FloatArgType, **kwargs: FloatArgType) -> float:  # type: ignore[misc]
        return float(*args, **kwargs)


class ErrorProxy(AnyAtomicType):
    """Class for XPath xs:error datatype. Accepts `None` or an empty sequence."""
    name = 'error'

    def __new__(cls, value: Any) -> list[NoReturn] | None:  # type: ignore[misc]
        if value is None or value == []:
            return cast(list[NoReturn] | None, value)
        msg = f"Cast {value!r} to xs:error is not possible"
        if isinstance(value, list):
            raise cls._invalid_value(msg)
        raise cls._invalid_type(msg)

    def __init__(self, value: Any) -> None:
        pass



#
# Copyright (c), 2018-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
from typing import cast, Union

from elementpath.aliases import NamespacesType, NsmapType
from elementpath.helpers import Patterns

# Namespaces
XML_NAMESPACE = "http://www.w3.org/XML/1998/namespace"
XMLNS_NAMESPACE = "http://www.w3.org/2000/xmlns/"  # Used in DOM for xmlns declarations
XSD_NAMESPACE = "http://www.w3.org/2001/XMLSchema"
XSI_NAMESPACE = "http://www.w3.org/2001/XMLSchema-instance"
XLINK_NAMESPACE = "http://www.w3.org/1999/xlink"

# XPath/XQuery namespaces
XPATH_FUNCTIONS_NAMESPACE = "http://www.w3.org/2005/xpath-functions"
XQT_ERRORS_NAMESPACE = "http://www.w3.org/2005/xqt-errors"
XPATH_MATH_FUNCTIONS_NAMESPACE = "http://www.w3.org/2005/xpath-functions/math"
XPATH_MAP_FUNCTIONS_NAMESPACE = "http://www.w3.org/2005/xpath-functions/map"
XPATH_ARRAY_FUNCTIONS_NAMESPACE = "http://www.w3.org/2005/xpath-functions/array"
XSLT_XQUERY_SERIALIZATION_NAMESPACE = "http://www.w3.org/2010/xslt-xquery-serialization"

# XML namespace attributes
XML_BASE = '{%s}base' % XML_NAMESPACE
XML_LANG = '{%s}lang' % XML_NAMESPACE
XML_SPACE = '{%s}space' % XML_NAMESPACE
XML_ID = '{%s}id' % XML_NAMESPACE

# XML Schema Instance namespace attributes
XSI_TYPE = '{%s}type' % XSI_NAMESPACE
XSI_NIL = '{%s}nil' % XSI_NAMESPACE
XSI_SCHEMA_LOCATION = '{%s}schemaLocation' % XSI_NAMESPACE
XSI_NONS_SCHEMA_LOCATION = '{%s}schemaLocation' % XSI_NAMESPACE

# XML Schema tags (schema and types)
XSD_SCHEMA = '{%s}schema' % XSD_NAMESPACE
XSD_ANY_TYPE = '{%s}anyType' % XSD_NAMESPACE
XSD_ANY_SIMPLE_TYPE = '{%s}anySimpleType' % XSD_NAMESPACE
XSD_ANY_ATOMIC_TYPE = '{%s}anyAtomicType' % XSD_NAMESPACE
XSD_NOTATION = '{%s}NOTATION' % XSD_NAMESPACE
XSD_ID = '{%s}ID' % XSD_NAMESPACE
XSD_IDREF = '{%s}IDREF' % XSD_NAMESPACE
XSD_IDREFS = '{%s}IDREFS' % XSD_NAMESPACE

XSD_STRING = '{%s}string' % XSD_NAMESPACE
XSD_FLOAT = '{%s}float' % XSD_NAMESPACE
XSD_DOUBLE = '{%s}double' % XSD_NAMESPACE
XSD_DECIMAL = '{%s}decimal' % XSD_NAMESPACE
XSD_DATETIME_STAMP = '{%s}dateTimeStamp' % XSD_NAMESPACE

# XPath type labels defined in XSD namespace that are not XSD builtin types
XSD_UNTYPED = '{%s}untyped' % XSD_NAMESPACE
XSD_UNTYPED_ATOMIC = '{%s}untypedAtomic' % XSD_NAMESPACE
XSD_ERROR = '{%s}error' % XSD_NAMESPACE
XSD_NUMERIC = '{%s}numeric' % XSD_NAMESPACE


def get_namespace(name: str) -> str:
    try:
        return Patterns.namespace_uri.match(name).group(1)  # type: ignore[union-attr]
    except AttributeError:
        return ''


def split_expanded_name(name: str) -> tuple[str, str]:
    match = Patterns.expanded_name.match(name)
    if match is None:
        raise ValueError(f"{name!r} is not an expanded QName")
    namespace, local_name = match.groups()
    return namespace or '', local_name


def get_prefixed_name(name: str, namespaces: Union[NamespacesType, NsmapType]) -> str:
    """
    Get the prefixed form of a QName, using a namespace map.

    :param name: an extended QName or a local name or a prefixed QName.
    :param namespaces: a dictionary with a map from prefixes to namespace URIs.
    """
    try:
        if not name.startswith(('{', 'Q{')):
            return name
        elif name[0] == '{':
            ns_uri, local_name = name[1:].split('}')
        else:
            ns_uri, local_name = name[2:].split('}')
    except (ValueError, TypeError):
        raise ValueError(f"{name!r} is not a QName")

    for prefix, uri in sorted(namespaces.items(), reverse=True,
                              key=lambda x: x if x[0] is not None else ('', x[1])):
        if uri == ns_uri:
            return f'{prefix}:{local_name}' if prefix else local_name
    else:
        if ns_uri == XML_NAMESPACE:
            return f'xml:{local_name}'
        return name


def get_expanded_name(name: str, namespaces: Union[NamespacesType, NsmapType]) -> str:
    """
    Get the expanded form of a QName, using a namespace map.
    Local names are mapped to the default namespace.

    :param name: a prefixed QName or a local name or an extended QName.
    :param namespaces: a dictionary with a map from prefixes to namespace URIs.
    :return: the expanded format of a QName or a local name.
    """
    if not name or name.startswith('{'):
        return name
    elif name.startswith('Q{'):
        return name[1:]

    try:
        prefix, local_name = name.split(':')
    except ValueError:
        if ':' in name:
            raise ValueError(f"wrong format for prefixed QName {name!r}")
        elif '' in namespaces:
            uri = namespaces['']
        elif None in namespaces:
            uri = cast(NsmapType, namespaces)[None]  # lxml nsmap
        else:
            return name

        return f'{{{uri}}}{name}' if uri else name
    else:
        if not prefix or not local_name:
            raise ValueError(f"wrong format for reference name {name!r}")
        elif prefix == 'xml':
            return f'{{{XML_NAMESPACE}}}{local_name}'

        uri = namespaces[prefix]
        if not uri:
            raise ValueError(f"prefix {prefix!r} is mapped to an empty URI")
        return f'{{{uri}}}{local_name}'


def get_expanded_path(path: str, namespaces: Union[NamespacesType, NsmapType]) -> str:
    """
    Similar to get_expanded_name, but can be applied to the whole path.
    """
    def expand_chunk(chunk: str) -> None:
        p = 0
        for m in Patterns.unbound_qname.finditer(chunk):
            s, e = m.span()
            if p < s:
                if ':' in chunk[p:s]:
                    idx = pos + p + chunk[p:s].index(':')
                    raise ValueError(f'unexpected colon character at position {idx}')
                chunks.append(chunk[p:s])

            prefix, local_part = m.groups()
            if prefix:
                uri = namespaces[prefix]
                chunks.append(f'{{{uri}}}{local_part}')
            elif default_ns:
                uri = default_ns
                chunks.append(f'{{{uri}}}{local_part}')
            else:
                chunks.append(local_part)

            p = e
        else:
            if p < len(chunk):
                if ':' in chunk[p:]:
                    idx = pos + p + chunk[p:].index(':')
                    raise ValueError(f'unexpected colon character at position {idx}')
                chunks.append(chunk[p:])

    chunks: list[str] = []
    pos = 0
    if None in namespaces:
        default_ns = namespaces[None]  # type: ignore[index, unused-ignore]
    else:
        default_ns = namespaces.get('', '')

    for match in Patterns.unbound_expanded_name.finditer(path):
        start, end = match.span()
        if pos < start:
            expand_chunk(path[pos:start])

        if path[start] == 'Q':
            chunks.append(path[start+1:end])
        else:
            chunks.append(path[start:end])
        pos = end
    else:
        if pos < len(path):
            expand_chunk(path[pos:])

    return ''.join(chunks)



#
# Copyright (c), 2018-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
"""
XPath 3.0 implementation - part 2 (symbols, operators and expressions)
"""
from copy import copy
from collections.abc import Iterator
from typing import Any, cast, Union

import elementpath.aliases as ta
import elementpath.namespaces as ns
from elementpath.datatypes import QName
from elementpath.xpath_tokens import XPathToken, ValueToken, XPathFunction, \
    XPathMap, XPathArray

from .xpath30_parser import XPath30Parser

__all__ = ['XPath30Parser']

register = XPath30Parser.register
infix = XPath30Parser.infix
method = XPath30Parser.method

register(':=')

###
# Placeholder symbol (used also for optional occurrence)

XPath30Parser.unregister('?')
register('?', bases=(ValueToken,))


@method('?')
def nud__placeholder_symbol(self: ValueToken) -> ValueToken:
    return self


@method('?')
def evaluate__placeholder_symbol(self: ValueToken, context: ta.ContextType = None) -> ValueToken:
    return self


###
# Braced/expanded QName(s)

XPath30Parser.duplicate('{', 'Q{', pattern=r'Q\{')
XPath30Parser.unregister('{')
XPath30Parser.unregister('}')
register('{')
register('}', bp=100)


XPath30Parser.unregister('(')


@method(register('(', lbp=80, rpb=80, label='expression'))
def nud__parenthesized_expression(self: XPathToken) -> XPathToken:
    if self.parser.next_token.symbol != ')':
        self[:] = self.parser.expression(),
    self.parser.advance(')')
    return self


@method('(')
def led__parenthesized_expression(self: XPathToken, left: XPathToken) -> XPathToken:
    if left.symbol in ('(name)', 'Q{'):
        if left.value in self.parser.RESERVED_FUNCTION_NAMES:
            msg = f"{left.value!r} is not allowed as function name"
            raise left.error('XPST0003', msg)
        else:
            raise left.error('XPST0017', 'unknown function {!r}'.format(left.value))

    elif left.symbol == ':' and left[1].symbol == '(name)':
        if left[1].namespace == ns.XSD_NAMESPACE:
            msg = 'unknown constructor function {!r}'.format(left[1].value)
            raise left[1].error('XPST0017', msg)
        raise left.error('XPST0017', 'unknown function {!r}'.format(left.value))

    if self.parser.next_token.symbol != ')':
        self[:] = left, self.parser.expression()
    else:
        self[:] = left,
    self.parser.advance(')')
    return self


@method('(')
def evaluate__parenthesized_expression(self: XPathToken, context: ta.ContextType = None) \
        -> ta.ValueType | XPathToken:
    if not self:
        return []

    value = self[0].evaluate(context)

    if isinstance(value, list) and len(value) == 1:
        value = value[0]

    if len(self) > 1:
        if isinstance(value, XPathFunction):
            func: XPathFunction
            func = value
            tokens = self[1].get_argument_tokens()

            if any(x.symbol == '?' and not x for x in tokens):
                func.check_arguments_number(len(tokens))
                func = copy(func)
                func[:] = tokens
                func.to_partial_function()
                return func

            arguments: list[ta.ValueType] = [tk.evaluate(context) for tk in tokens]

            if func.label == 'partial function' and func[0].symbol == '?' and len(func[0]):
                if context is None:
                    raise self.missing_context()
                return func(context.item, *arguments, context=context)

            return func(*arguments, context=context)

        elif self[0].symbol == '(':
            if not isinstance(value, list):
                return value
            elif any(not isinstance(x, XPathFunction) for x in value):
                return value

        if isinstance(value, XPathToken) and value.symbol == '?':
            return value

        raise self.error('XPTY0004', f'an XPath function expected, not {type(value)!r}')

    if isinstance(value, (XPathMap, XPathArray)) or \
            not isinstance(value, XPathFunction) or self[0].span[0] > self.span[0]:
        return value
    else:
        return value(context=context)


@method(infix('||', bp=32))
def evaluate__union_operator(self: XPathToken, context: ta.ContextType = None) -> str:

    return self.string_value(self.get_argument(context)) + \
        self.string_value(self.get_argument(context, index=1))


@method(infix('!', bp=72))
def select__simple_map_operator(self: XPathToken, context: ta.ContextType = None) \
        -> Iterator[ta.ItemType]:
    if context is None:
        raise self.missing_context()

    for context.item in self[0].select_with_focus(context):
        for result in self[1].select(context):
            yield result


###
# 'let' expressions

@method('let', bp=20, label='expression')
def nud__let_expression(self: XPathToken) -> XPathToken:
    del self[:]
    if self.parser.next_token.symbol != '$':
        return self.as_name()

    while True:
        self.parser.next_token.expected('$')
        variable = self.parser.expression(5)
        self.append(variable)
        self.parser.advance(':=')
        expr = self.parser.expression(5)
        self.append(expr)
        if self.parser.next_token.symbol != ',':
            break
        self.parser.advance()

    self.parser.advance('return')
    self.append(self.parser.expression(5))
    return self


@method('let')
def select__let_expression(self: XPathToken, context: ta.ContextType = None) \
        -> Iterator[ta.ItemType]:
    if context is None:
        raise self.missing_context()

    context = copy(context)
    context.variables = context.variables.copy()

    for k in range(0, len(self) - 1, 2):
        varname = cast(str, self[k][0].value)
        value = self[k+1].evaluate(context)
        context.variables[varname] = value

    yield from self[-1].select(context)


@method('#', bp=90)
def led__function_reference(self: XPathToken, left: XPathToken) -> XPathToken:
    if not left.label.endswith('function'):
        left.expected(':', '(name)', 'Q{')

    self[:] = left, self.parser.expression(rbp=90)
    self[1].expected('(integer)')
    return self


@method('#')
def evaluate__function_reference(self: XPathToken, context: ta.ContextType = None) -> XPathFunction:
    token_class: type[Union[XPathFunction, XPathToken]]
    namespace: Any
    name: Any

    arity = self[1].value
    assert arity is None or isinstance(arity, int)

    if isinstance(self[0], XPathFunction):
        token_class = self[0].__class__
        namespace = self[0].namespace
        name = self[0].name
        if isinstance(name, QName):
            qname = name
        else:
            qname = QName(None, f'anonymous {self[0].label}'.replace(' ', '-'))
    else:
        if self[0].symbol == ':':
            namespace = self[0][1].namespace
            name = self[0].value
        elif self[0].symbol == 'Q{':
            namespace = self[0][0].value
            name = self[0][1].value
        elif self[0].value not in self.parser.RESERVED_FUNCTION_NAMES:
            namespace = ns.XPATH_FUNCTIONS_NAMESPACE
            name = self[0].value
        else:
            msg = f"{self[0].value!r} is not allowed as function name"
            raise self.error('XPST0003', msg)

        assert isinstance(name, str)
        assert isinstance(namespace, str) or namespace is None
        qname = QName(namespace, name)
        namespace = qname.namespace
        local_name = qname.local_name

        # Generic rule for XSD constructor functions
        if namespace == ns.XSD_NAMESPACE and arity != 1:
            raise self.error('XPST0017', f"unknown function {qname.qname}#{arity}")

        # Special checks for multirole tokens
        if namespace == ns.XPATH_FUNCTIONS_NAMESPACE and \
                local_name in ('QName', 'dateTime') and arity == 1:
            raise self.error('XPST0017', f"unknown function {qname.qname}#{arity}")

        try:
            token_class = self.parser.symbol_table[qname.expanded_name]
        except KeyError:
            try:
                token_class = self.parser.symbol_table[local_name]
            except KeyError:
                msg = f"unknown function {qname.qname}#{arity}"
                raise self.error('XPST0017', msg) from None

        if token_class.symbol == 'function' or not token_class.label.endswith('function'):
            raise self.error('XPST0003')
        assert issubclass(token_class, XPathFunction)

    try:
        func = token_class(self.parser, nargs=arity)
    except TypeError:
        msg = f"unknown function {qname.qname}#{arity}"
        raise self.error('XPST0017', msg) from None
    else:
        if func.namespace is None:
            func.namespace = namespace
        elif func.namespace != namespace:
            raise self.error('XPST0017', f"unknown function {qname.qname}#{arity}")
        func.context = copy(context)
        return func



#
# Copyright (c), 2022-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
import math
from decimal import Decimal
from functools import cmp_to_key
from itertools import zip_longest
from typing import Any, Optional

from collections.abc import Callable, Iterable, Iterator
from elementpath.protocols import ElementProtocol
from elementpath.exceptions import xpath_error
from elementpath.datatypes import UntypedAtomic, AnyURI, AbstractQName
from elementpath.collations import UNICODE_CODEPOINT_COLLATION, CollationManager
from elementpath.xpath_nodes import XPathNode, EtreeElementNode, TextAttributeNode, \
    NamespaceNode, TextNode, CommentNode, ProcessingInstructionNode, EtreeDocumentNode
from elementpath.xpath_tokens import XPathToken, XPathFunction, XPathMap, XPathArray


def deep_equal(seq1: Iterable[Any],
               seq2: Iterable[Any],
               collation: Optional[str] = None,
               token: Optional[XPathToken] = None) -> bool:

    etree_node_types = (EtreeElementNode, CommentNode, ProcessingInstructionNode)

    def etree_deep_equal(e1: ElementProtocol, e2: ElementProtocol) -> bool:
        if cm.ne(e1.tag, e2.tag):
            return False
        elif cm.ne((e1.text or '').strip(), (e2.text or '').strip()):
            return False
        elif cm.ne((e1.tail or '').strip(), (e2.tail or '').strip()):
            return False
        elif len(e1) != len(e2) or len(e1.attrib) != len(e2.attrib):
            return False

        try:
            items1 = {(cm.strxfrm(k or ''), cm.strxfrm(v))  # type: ignore[arg-type]
                      for k, v in e1.attrib.items()}
            items2 = {(cm.strxfrm(k or ''), cm.strxfrm(v))  # type: ignore[arg-type]
                      for k, v in e2.attrib.items()}
        except TypeError:
            return False

        if items1 != items2:
            return False
        return all(etree_deep_equal(c1, c2) for c1, c2 in zip(e1, e2))

    if collation is None:
        collation = UNICODE_CODEPOINT_COLLATION

    with CollationManager(collation, token=token) as cm:
        for value1, value2 in zip_longest(seq1, seq2):
            if isinstance(value1, XPathFunction) and \
                    not isinstance(value1, (XPathMap, XPathArray)):
                raise xpath_error('FOTY0015', token=token)
            if isinstance(value2, XPathFunction) and \
                    not isinstance(value2, (XPathMap, XPathArray)):
                raise xpath_error('FOTY0015', token=token)

            if (value1 is None) ^ (value2 is None):
                return False
            elif isinstance(value1, (XPathArray, XPathMap, XPathNode)) ^ \
                    isinstance(value2, (XPathArray, XPathMap, XPathNode)):
                return False
            elif value1 is None:
                return True
            elif isinstance(value1, XPathMap):
                assert isinstance(value2, XPathMap)
                return value1 == value2
            elif isinstance(value1, XPathArray):
                assert isinstance(value2, XPathArray)
                return value1 == value2
            elif isinstance(value1, XPathNode):
                assert isinstance(value2, XPathNode)
                if value1.__class__ != value2.__class__:
                    return False
                elif isinstance(value1, etree_node_types):
                    assert isinstance(value2, etree_node_types)
                    if not etree_deep_equal(value1.value, value2.value):
                        return False
                elif isinstance(value1, EtreeDocumentNode):
                    assert isinstance(value2, EtreeDocumentNode)
                    for child1, child2 in zip_longest(value1, value2):
                        if child1 is None or child2 is None:
                            return False
                        elif child1.__class__ != child2.__class__:
                            return False
                        elif isinstance(child1, etree_node_types):
                            assert isinstance(child2, etree_node_types)
                            if not etree_deep_equal(child1.value, child2.value):
                                return False
                        elif isinstance(child1, TextNode):
                            assert isinstance(child2, TextNode)
                            if cm.ne(child1.value, child2.value):
                                return False

                elif cm.ne(value1.value, value2.value):
                    return False
                elif isinstance(value1, TextAttributeNode):
                    if cm.ne(value1.name, value2.name):
                        return False
                elif isinstance(value1, NamespaceNode):
                    assert isinstance(value2, NamespaceNode)
                    if cm.ne(value1.prefix, value2.prefix):
                        return False
            else:
                try:
                    if isinstance(value1, bool):
                        if not isinstance(value2, bool) or value1 is not value2:
                            return False

                    elif isinstance(value2, bool):
                        return False

                    if isinstance(value1, AbstractQName):
                        if not isinstance(value2, AbstractQName) or value1 != value2:
                            return False

                    elif isinstance(value2, AbstractQName):
                        return False

                    elif isinstance(value1, (str, AnyURI, UntypedAtomic)) \
                            and isinstance(value2, (str, AnyURI, UntypedAtomic)):
                        if cm.strcoll(str(value1), str(value2)):
                            return False

                    elif isinstance(value1, UntypedAtomic) \
                            or isinstance(value2, UntypedAtomic):
                        return False

                    elif isinstance(value1, float):
                        if math.isnan(value1):
                            if not math.isnan(value2):
                                return False
                        elif math.isinf(value1):
                            if value1 != value2:
                                return False
                        elif isinstance(value2, Decimal):
                            if value1 != float(value2):
                                return False
                        elif not isinstance(value2, (value1.__class__, int)):
                            return False
                        elif value1 != value2:
                            return False

                    elif isinstance(value2, float):
                        if math.isnan(value2):
                            return False
                        elif math.isinf(value2):
                            if value1 != value2:
                                return False
                        elif isinstance(value1, Decimal):
                            if value2 != float(value1):
                                return False
                        elif not isinstance(value1, (value2.__class__, int)):
                            return False
                        elif value1 != value2:
                            return False
                    elif value1 != value2:
                        return False
                except TypeError:
                    return False

    return True


def deep_compare(obj1: Any,
                 obj2: Any,
                 collation: Optional[str] = None,
                 token: Optional[XPathToken] = None) -> int:

    msg_tmpl = "Sorting failed, cannot compare {!r} with {!r}"
    etree_node_types = (EtreeElementNode, CommentNode, ProcessingInstructionNode)
    result: int = 0

    def iter_object(obj: Any) -> Iterator[Any]:
        if isinstance(obj, XPathArray):
            yield from obj.items()
        elif isinstance(obj, list):
            yield from obj
        else:
            yield obj

    def etree_deep_compare(e1: ElementProtocol, e2: ElementProtocol) -> int:
        nonlocal result
        if not callable(e1.tag) and not callable(e2.tag):
            result = cm.strcoll(e1.tag, e2.tag)
            if result:
                return result

        result = cm.strcoll((e1.text or '').strip(), (e2.text or '').strip())
        if result:
            return result

        for a1, a2 in zip_longest(e1.attrib.items(), e2.attrib.items()):
            if a1 is None:
                return 1
            elif a2 is None:
                return -1

            result = cm.strcoll(a1[0], a2[0]) or cm.strcoll(a1[1], a2[1])
            if result:
                return result

        for c1, c2 in zip_longest(e1, e2):
            if c1 is None:
                return 1
            elif c2 is None:
                return -1

            result = etree_deep_compare(c1, c2)
            if result:
                return result
        else:
            result = cm.strcoll((e1.tail or '').strip(), (e2.tail or '').strip())
            if result:
                return result

            return 0

    if collation is None:
        collation = UNICODE_CODEPOINT_COLLATION

    with CollationManager(collation, token=token) as cm:
        for value1, value2 in zip_longest(iter_object(obj1), iter_object(obj2)):
            if isinstance(value1, XPathFunction) and \
                    not isinstance(value1, XPathArray):
                raise xpath_error('FOTY0013', token=token)
            if isinstance(value2, XPathFunction) and \
                    not isinstance(value2, XPathArray):
                raise xpath_error('FOTY0013', token=token)

            if value1 is None or value1 == []:
                if value2 is not None and value2 != []:
                    return -1

            if value2 is None or value2 == []:
                if value1 is not None and value1 != []:
                    return 1

            if isinstance(value1, XPathNode) ^ isinstance(value2, XPathNode):
                msg = f"cannot compare {type(value1)} with {type(value2)}"
                raise xpath_error('XPTY0004', msg, token=token)
            elif isinstance(value1, XPathNode):
                assert isinstance(value2, XPathNode)
                if value1.__class__ != value2.__class__:
                    msg = f"cannot compare {type(value1)} with {type(value2)}"
                    raise xpath_error('XPTY0004', msg, token=token)
                elif isinstance(value1, etree_node_types):
                    assert isinstance(value2, etree_node_types)
                    result = etree_deep_compare(value1.value, value2.value)
                    if result:
                        return result
                elif isinstance(value1, EtreeDocumentNode):
                    assert isinstance(value2, EtreeDocumentNode)
                    for child1, child2 in zip_longest(value1, value2):
                        if child1 is None:
                            return -1
                        elif child2 is None:
                            return 1
                        elif child1.__class__ != child2.__class__:
                            msg = f"cannot compare {type(child1)} with {type(child2)}"
                            raise xpath_error('XPTY0004', msg, token=token)
                        elif isinstance(child1, etree_node_types):
                            assert isinstance(child2, etree_node_types)
                            result = etree_deep_compare(child1.value, child2.value)
                            if result:
                                return result
                        elif isinstance(child1, TextNode):
                            assert isinstance(child2, TextNode)
                            result = cm.strcoll(
                                child1.value.strip(), child2.value.strip()
                            )
                            if result:
                                return result
                elif isinstance(value1, TextNode):
                    assert isinstance(value2, TextNode)
                    result = cm.strcoll(value1.value, value2.value)
                    if result:
                        return result
                elif isinstance(value1, TextAttributeNode):
                    assert isinstance(value2, TextAttributeNode)
                    result = cm.strcoll(value1.name or '', value2.name or '')
                    if result:
                        return result
                elif isinstance(value1, NamespaceNode):
                    assert isinstance(value2, NamespaceNode)
                    result = cm.strcoll(value1.prefix or '', value2.prefix or '')
                    if result:
                        return result
            else:
                try:
                    if isinstance(value1, bool):
                        if not isinstance(value2, bool):
                            return -1
                        elif value1 is not value2:
                            return -1 if value1 else 1

                    elif isinstance(value2, bool):
                        return -1

                    elif isinstance(value1, UntypedAtomic):
                        if isinstance(value2, UntypedAtomic):
                            result = cm.strcoll(str(value1), str(value2))
                            if result:
                                return result
                        else:
                            msg = msg_tmpl.format(value1, value2)
                            raise xpath_error('XPTY0004', msg, token)

                    elif isinstance(value2, UntypedAtomic):
                        msg = msg_tmpl.format(value1, value2)
                        raise xpath_error('XPTY0004', msg, token)

                    elif isinstance(value1, float):
                        if math.isnan(value1):
                            if not isinstance(value2, (float, Decimal)) \
                                    or not math.isnan(value2):
                                return -1
                        elif math.isinf(value1):
                            if value1 != value2:
                                return -1 if value1 < value2 else 1
                        elif isinstance(value2, Decimal):
                            if value1 != float(value2):
                                return -1 if value1 < float(value2) else 1
                        elif not isinstance(value2, (value1.__class__, int)):
                            return -1
                        elif value1 != value2:
                            return -1 if value1 < value2 else 1

                    elif isinstance(value2, float):
                        if math.isnan(value2):
                            return -1
                        elif math.isinf(value2):
                            if value1 != value2:
                                return -1 if value1 < value2 else 1
                        elif isinstance(value1, Decimal):
                            if value2 != float(value1):
                                return -1 if float(value1) < value2 else 1
                        elif not isinstance(value1, (value2.__class__, int)):
                            return -1
                        elif value1 != value2:
                            return -1 if value1 < value2 else 1

                    elif isinstance(value1, (str, AnyURI, UntypedAtomic)) \
                            and isinstance(value2, (str, AnyURI, UntypedAtomic)):
                        result = cm.strcoll(str(value1), str(value2))
                        if result:
                            return result
                    elif value1 != value2:
                        return -1 if value1 < value2 else 1

                except TypeError as err:
                    raise xpath_error('XPTY0004', message_or_error=err, token=token)

    return 0


def get_key_function(collation: Optional[str] = None,
                     key_func: Optional[Callable[[Any], Any]] = None,
                     token: Optional[XPathToken] = None) -> Any:

    def compare_func(obj1: Any, obj2: Any) -> int:
        if key_func is not None:
            if isinstance(obj1, list):
                obj1 = map(key_func, obj1)
            else:
                obj1 = key_func(obj1)

            if isinstance(obj2, list):
                obj2 = map(key_func, obj2)
            else:
                obj2 = key_func(obj2)

        return deep_compare(obj1, obj2, collation, token)

    return cmp_to_key(compare_func)


def same_key(k1: Any, k2: Any) -> bool:
    if isinstance(k1, (str, AnyURI, UntypedAtomic)):
        if not isinstance(k2, (str, AnyURI, UntypedAtomic)):
            return False
        return str(k1) == str(k2)
    elif isinstance(k1, float) and math.isnan(k1):
        return isinstance(k2, float) and math.isnan(k2)
    elif isinstance(k1, AbstractQName) ^ isinstance(k2, AbstractQName):
        return False

    try:
        return True if k1 == k2 else False
    except TypeError:
        return False  # EAFP :)
